#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from pandas import to_datetime
import datetime
import argparse
import sys


def get_datetime_from_args(date_time):
    """
    Datetime strings from the commandline can come in a list, so we need to
    join it back in as a single string for pandas to handle it.
    """
    if type(date_time) == list:
        date_time = " ".join(date_time)

    if type(date_time) == str:
        date_time = to_datetime(date_time)
    else:
        date_time = date_time

    return date_time


def handle_year_stradling(date_time):
    """
    Deals with the scenario of the water year starting in
    10/1/N  and ending in 10/1/N+1, assumes that the water year is asssociated
    to the bulk of time at the year which is N+1

    Depending on the time, will return the appropiate year for the starting
    point
    """

    if date_time.month < 10:
        year = date_time.year-1
    else:
        year = date_time.year

    return year

def calculate_end_of_day(date_time):
    """
    Calculates the end of the day given date
    """
    date_time = datetime.datetime(year=date_time.year, month=date_time.month,
                                                   day=date_time.day,
                                                   hour=23)
    return date_time


def calculate_date_from_wyhr(wyhr, year):
    """
    Takes in the integer of water year hours and an integer year and
    returns the date
    """

    start = datetime.datetime(year=year-1, month=10, day=1)

    delta = datetime.timedelta(hours = wyhr)
    return start+delta


def calculate_wyhr_from_date(date_time):
    """
    Takes in the string of the datetime and returns an integer of the water year
    hour.

    Args:

        date_time: string or date time objec to use for calculating the water year
    """

    year = handle_year_stradling(date_time)

    start = datetime.datetime(year=year, month=10, day=1)

    delta = date_time - start

    return int((delta.total_seconds())/3600)


def main():
    parser = argparse.ArgumentParser(description="Converts dates and water year hour")
    parser.add_argument('convertable', nargs='+', default=None,
                        help="Receives either a datetime or wyhr and "
                        "infers automatically")

    parser.add_argument("--date","-d", type=str, dest='date',nargs='+',
                    help="String of date time to convert to water year hour")
    parser.add_argument('--wyhr', dest='wyhr', type=int,
                        help="Integer of water year hours")
    parser.add_argument('--year','-y', dest='year', type=int,
                        help="Integer of year, required when converting from"
                             " wyhr")
    parser.add_argument('--eod', dest='eod', action='store_true',
                        help="Flag for calculating the end of the day wyhr")
    args = parser.parse_args()

    # Snag todays date
    today = datetime.datetime.now()

    # Print a header
    msg = "WYHR Conversion Utility:"
    hdr = "=" * len(msg)
    print("\n")
    print(msg)
    print(hdr)

    # Try to infer the users input if no args are provided.
    if args.convertable != None:

        # Check for wyhr
        try:
            args.wyhr = int(" ".join(args.convertable))

        except:
            args.wyhr = None

        # Check for datetime
        if args.wyhr == None:
            try:
                date = get_datetime_from_args(" ".join(args.convertable))
                args.date = args.convertable

            except Exception as e:
                print("\nERROR: Could not distinguish argument as a WYHR or a"
                      " datetime.\nException: {}\n".format(e))
                sys.exit()

    # Convert from date specified
    if args.date != None:
        date = get_datetime_from_args(args.date)
        from_date = True

    # Convert from water year hour
    elif args.wyhr != None and args.year != None :
        wyhr = args.wyhr
        year=args.year
        from_date = False

    # Assume water year conversion from this year
    elif args.wyhr != None and args.year == None:
        wyhr = args.wyhr
        year = today.year
        from_date = False

    # Assume calculate wyhr from today
    else:
        date = today
        from_date = True

    # Perform the final calculation
    if from_date:
        # Force the end of the day
        if args.eod:
            date = calculate_end_of_day(date)

        wyhr = calculate_wyhr_from_date(date)

    else:
        date = calculate_date_from_wyhr(wyhr, year)


    print("Date: {0}".format(date.strftime("%Y-%m-%d %H:00")))
    print("WYHR: {0}".format(wyhr))
    print('\n')

if __name__ == '__main__':
    main()




import os
import netCDF4 as nc

def get_topo_stats(path):
    """
    Get stats about topo from the topo file.

    Returns
    ------
    data - dictionary of topo header data
    """

    path = os.path.abspath(path)
    data = nc.Dataset(path, 'r')

    stats = {}
    stats['y'] = data.variables['y'][:]
    stats['x'] = data.variables['x'][:]
    stats['nx'] = len(stats['x'])
    stats['ny'] = len(stats['y'])
    stats['du'] = stats['y'][1] - stats['y'][0]
    stats['dv'] = stats['x'][1] - stats['x'][0]
    stats['v'] = stats['x'][0]
    stats['u'] = stats['y'][0]

    data.close()

    return stats




def plotlims(plotorder):
    '''
    Compile basin-specific limits for making tidy figures.

    Args
    ------
    plotorder : list
        subbasin names

    Returns
    ------
    lims : class
        basin-specific plot details:
        pbbx:
        legx: legend x
        legy: legend y
        btx: basin total text x
        bty: basin total text y
        sumorder: list of subbasin to plot (omits total unless single basin)
        swid: bar spacing

    '''
    basin = plotorder[0]

    class lims:

        # snow-free patch label
        pbbx = 0.05

        # legend x,y
        legx = 0.01
        legy = 0.71

        # second legend x,y for precip_depth
        legy2 = 0.71

        btx = 0.31
        bty = 0.95

        if 'San Joaquin' in basin:
            pbbx = 0.3
            legy = 0.69
            legy2 = 0.65
            basin_name = 'sanjoaquin'

        if basin in ['RCEW']:
            pbbx = 0.1

        if 'Kings' in basin:
            legy = 0.52
            legy2 = 0.39
            basin_name = 'kings'

        if 'Merced' in basin:
            legy = 0.58
            legy2 = 0.42
            basin_name = 'merced'

        if 'Kaweah' in basin:
            legy = 0.52
            legy2 = 0.38
            pbbx = 0.45
            basin_name = 'kaweah'

        if 'Lakes' in basin:
            btx = 0.26
            legy = 0.85
            legy2 = 0.88
            basin_name = 'lakes'

        if basin == 'Extended Tuolumne':
            legy = 0.75
            btx = 0.28
            basin_name = 'tuolumne'

        if basin == 'Tuolumne River Basin':
            legy = 0.46
            btx = 0.28
            pbbx = 0.25
            legy2 = 0.30
            basin_name = 'tuolumne'

        if 'Boise' in basin:
            legy = 0.75
            btx = 0.28
            basin_name = 'brb'

        # Total basin label
        if plotorder is not None:
            if len(plotorder) > 1:
                sumorder = plotorder[1::]
                swid = 0.25

            else:
                sumorder = plotorder
                swid = 0.45

    return lims




from . import basin_total
from . import cold_content
from . import compare_runs
from . import density
from . import flt_image_change
from . import image_change
from . import inputs
from . import plotlims
from . import point_values
from . import precip_depth
from . import swe_volume
from . import stn_validate
from . import swi
from . import write_properties



import copy
import cmocean
from datetime import timedelta
from matplotlib import pyplot as plt
import matplotlib.colors as mcolors
from mpl_toolkits.axes_grid1 import make_axes_locatable
import netCDF4 as nc
import numpy as np
import os
import seaborn as sns

from snowav.utils.wyhr import calculate_date_from_wyhr
from snowav.utils.MidpointNormalize import MidpointNormalize
from snowav.database.database import collect
import snowav.framework.figures


def flt_image_change(file, update_numbers, end_date, flight_outputs,
                     pre_flight_outputs, masks, lims, barcolors, edges,
                     connector, plotorder, wy, depth_factor, basins,
                     run_name, figsize, depthlbl, elevlbl, vollbl, dplcs,
                     figspath, dpi=200, logger=None):
    """ Difference in SWE from one day prior to flight updates to the day of
    flight updates.

    The file specified in [plots] update_file is used to get dates of flight
    updates and the masks for which portions of the basin were updated. The
    actual SWE differences are between the snow.nc files on the dates of the
    flights and the day immediately prior. [plots] update_numbers is a 1-based
    list of subsetting flights that can be applied.

    Args
    ------
    file {str}: path to lidar_depths.nc
    update_numbers {list}: flight numbers to generate
    end_date {datetime}: end date
    flight_outputs {dict}: from config.py
    pre_flight_outputs {dict}: from config.py
    masks {dict}: from config.py
    lims {list}: percent of min/max for color limits
    barcolors {list}: colors
    edges {list}: elevation bands
    connector {str}: database connector
    plotorder {list}: basins
    wy {int}: water year
    depth_factor {float}: depth factor
    basins {dict}: from config.py
    run_name {str}: snowav run_name
    figsize {list}: figure dimensions
    depthlbl {str}: depth label
    elevlbl {str}: elevation label
    vollbl {str}: volume label
    dplcs {int}: decimal places
    figspath {str}: base path for figures
    dpi {int}: figure dpi
    """

    e = False
    basin_name = plotorder[0].lower().split(" ")[0]
    flight_diff_fig_names = []
    flight_delta_vol_df = {}

    p = nc.Dataset(file, 'r')

    if update_numbers is None:
        times = p.variables['time'][:]
    else:
        if max(update_numbers) > len(p.variables['time']) - 1:
            raise Exception('Max flight update_numbers greater than available '
                            'flights')
        else:
            times = p.variables['time'][update_numbers]

    # remove flight indices that might be after the report period
    idx = []

    for i, n in enumerate(times):
        date = calculate_date_from_wyhr(int(n), wy)
        if date > end_date:
            idx = np.append(idx, i)

    times = np.delete(times, idx)
    ix = np.argsort(times)
    times = times[ix]

    if update_numbers is not None:
        update_numbers = [int(x) for x in update_numbers]

    for i, time in enumerate(times):
        if update_numbers is not None:
            depth = p.variables['depth'][update_numbers[ix[i]], :, :]
        else:
            depth = p.variables['depth'][ix[i], :, :]

        delta_swe = flight_outputs['swe_z'][i][:] - pre_flight_outputs['swe_z'][i][:]
        delta_swe = delta_swe * depth_factor

        # We will always make the difference between the previous day
        start_date = flight_outputs['dates'][i] - timedelta(hours=24)
        end_date = flight_outputs['dates'][i]

        try:
            delta_swe_byelev = collect(connector, plotorder, basins, start_date,
                                       end_date, 'swe_vol', run_name, edges,
                                       'difference')
        except:
            e = True

        try:
            end_swe_byelev = collect(connector, plotorder, basins, start_date,
                                     end_date, 'swe_vol', run_name, edges,
                                     'end')
        except:
            e = True

        if e:
            if logger is not None:
                logger.info(' Failed requesting database records ending on {} '
                            'for flight difference figure. This may mean that '
                            '[run] directory has not been processed with '
                            '[snowav] run_name: {} for the periods in '
                            '[plots] update_file. Try subsetting with [plots] '
                            'update_numbers or processing the full [run] '
                            'directory.'.format(end_date, run_name))
                logger.info(' Flight figures being set to False...')

            return [], []

        # Make copy so that we can add nans for the plots
        delta_state = copy.deepcopy(delta_swe)
        v_min, v_max = np.nanpercentile(delta_state, [lims[0], lims[1]])

        colorsbad = plt.cm.Set1_r(np.linspace(0., 1, 1))
        colors1 = cmocean.cm.matter_r(np.linspace(0., 1, 127))
        colors2 = plt.cm.Blues(np.linspace(0, 1, 128))
        colors = np.vstack((colorsbad, colors1, colors2))
        mymap = mcolors.LinearSegmentedColormap.from_list('my_colormap', colors)

        ixf = delta_state == 0
        delta_state[ixf] = -100000
        pmask = masks[plotorder[0]]['mask']
        ixo = pmask == 0
        delta_state[ixo] = np.nan
        cmap = copy.copy(mymap)
        cmap.set_bad('white', 1.)

        sns.set_style('darkgrid')
        sns.set_context("notebook")

        plt.close(i)
        fig, (ax, ax1) = plt.subplots(num=i, figsize=figsize, dpi=dpi, nrows=1,
                                      ncols=2)

        norm = MidpointNormalize(midpoint=0, vmin=v_min, vmax=v_max)

        # this is primarily for awsm_test_cases, in which the flight update
        # nc file doesn't contain mask information. wy2019-forward flight nc
        # files *should* have a mask
        if hasattr(depth, 'mask'):
            mask = np.ma.masked_array(depth.mask, ~depth.mask)
            if mask.shape != delta_state.shape:
                raise Exception('Dimensions {}: {} do not match snow.nc: '
                                '{}'.format(file, mask.shape, delta_state.shape))

            delta_state[mask] = np.nan
            h = ax.imshow(delta_state * mask, cmap=cmap, clim=(v_min, v_max),
                          norm=norm)

        else:
            h = ax.imshow(delta_state, cmap=cmap, clim=(v_min, v_max),
                          norm=norm)

        for name in masks:
            ax.contour(masks[name]['mask'], cmap="Greys", linewidths=1)

        h.axes.get_xaxis().set_ticks([])
        h.axes.get_yaxis().set_ticks([])
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.2)
        cbar = plt.colorbar(h, cax=cax)
        cbar.set_label(r'$\Delta$ SWE [{}]'.format(depthlbl))

        start_date = start_date + timedelta(hours=1)
        d = end_date + timedelta(hours=1)

        h.axes.set_title('Change in SWE Depth from Update\n{}'
                         .format(start_date.date().strftime("%Y-%-m-%-d")))

        if len(plotorder) == 1:
            porder = plotorder
        else:
            porder = plotorder[1::]

        for iters, name in enumerate(porder):
            if dplcs == 0:
                lbl = '{}: {} {}'.format(name,
                                         str(int(delta_swe_byelev[name].sum())),
                                         vollbl)
            else:
                lbl = '{}: {} {}'.format(name,
                                         str(np.round(delta_swe_byelev[name].sum(),
                                                      dplcs)), vollbl)

            if iters == 0:
                ax1.bar(range(0, len(edges)), delta_swe_byelev[name],
                        color=barcolors[iters],
                        edgecolor='k', label=lbl)

            else:
                ax1.bar(range(0, len(edges)), delta_swe_byelev[name],
                        color=barcolors[iters],
                        edgecolor='k',
                        label=lbl,
                        alpha=0.5)

        end_swe_byelev = end_swe_byelev.fillna(0)
        datestr = flight_outputs['dates'][i].date().strftime("%Y%m%d")
        percent_delta_byelev = (delta_swe_byelev.sum(skipna=True) / end_swe_byelev.sum()) * 100
        flight_delta_vol_df[flight_outputs['dates'][i].date().strftime("%Y%m%d")] = delta_swe_byelev

        fp = os.path.join(figspath,
                          '{}_flight_{}_delta_taf.csv'.format(basin_name, datestr))
        delta_swe_byelev.to_csv(fp)
        fp = os.path.join(figspath,
                          '{}_flight_{}_percent_change.csv'.format(basin_name, datestr))
        percent_delta_byelev.to_csv(fp)

        plt.tight_layout()
        xts = ax1.get_xticks()
        edges_lbl = []
        for x in xts[0:len(xts) - 1]:
            edges_lbl.append(str(int(edges[int(x)])))

        ax1.set_xticklabels(str(x) for x in edges_lbl)
        for tick in ax1.get_xticklabels():
            tick.set_rotation(30)

        ylims = ax1.get_ylim()
        ymax = ylims[1] + abs(ylims[1] - ylims[0])
        ymin = ylims[0] - abs(ylims[1] - ylims[0]) * 0.1
        ax1.set_ylim((ymin, ymax))

        ax1.set_ylabel(r'$\Delta$ SWE [{}]'.format(vollbl))
        ax1.set_xlabel('elevation [{}]'.format(elevlbl))

        ax1.yaxis.set_label_position("right")
        ax1.tick_params(axis='x')
        ax1.tick_params(axis='y')
        ax1.yaxis.tick_right()

        if len(plotorder) > 1:
            ax1.legend(loc=2, fontsize=10)

        plt.tight_layout()
        fig.subplots_adjust(top=0.88)

        fstr = '{}_flight_difference_{}.png'.format(basin_name, datestr)
        flight_diff_fig_names += [fstr]
        fig_path = os.path.join(figspath, fstr)

        snowav.framework.figures.save_fig(fig, fig_path)

        if logger is not None:
            logger.info(' Saved: {}'.format(fig_path))

    p.close()

    return flight_diff_fig_names, flight_delta_vol_df




from snowav.utils.MidpointNormalize import MidpointNormalize
import numpy as np
from matplotlib import pyplot as plt
import matplotlib.colors as mcolors
from mpl_toolkits.axes_grid1 import make_axes_locatable
import seaborn as sns
import cmocean
import matplotlib.patches as mpatches
import pandas as pd
import copy
import snowav.framework.figures

def precip_depth(args, logger = None):
    '''
    Depth of SWI, precipitation, and rain during the report period.

    Args
    ----------
    args : dict
        dictionary with required inputs, see swi() figure for more information.

    logger : list
        snowav logger

    '''

    accum = args['swi_image']
    precip = args['precip_image']
    rain = args['rain_image']
    accum_byelev = args['swi_df']
    precip_byelev = args['precip_df']
    rain_byelev = args['rain_df']
    plotorder = args['plotorder']
    barcolors = args['barcolors']
    lims = args['lims']
    masks = args['masks']
    edges = args['edges']
    labels = args['labels']

    mm = [args['percent_min'],args['percent_max']]

    if np.nanpercentile(accum,mm)[1] > np.nanpercentile(precip,mm)[1]:
        z,qMax = np.nanpercentile(accum,mm)
    else:
        z,qMax = np.nanpercentile(precip,mm)

    if np.nanpercentile(accum,mm)[0] < np.nanpercentile(precip,mm)[0]:
        qMin,z = np.nanpercentile(accum,mm)
    else:
        qMin,z = np.nanpercentile(precip,mm)

    clims = (0,qMax)

    # Get bar plot ylims
    if np.nanmax(accum_byelev.values) > np.nanmax(precip_byelev.values):
        if len(plotorder) < 5:
            yMax = np.nanmax(accum_byelev.values) + np.nanmax(accum_byelev.values)*0.4
        else:
            yMax = np.nanmax(accum_byelev.values) + np.nanmax(accum_byelev.values)*0.6
    else:
        if len(plotorder) < 5:
            yMax = np.nanmax(precip_byelev.values) + np.nanmax(precip_byelev.values)*0.4
        else:
            yMax = np.nanmax(precip_byelev.values) + np.nanmax(precip_byelev.values)*0.6

    colors1 = cmocean.cm.dense(np.linspace(0, 1, 255))
    colors2 = plt.cm.Set1_r(np.linspace(0, 1, 1))
    colors = np.vstack((colors2, colors1))
    cmap = mcolors.LinearSegmentedColormap.from_list('my_colormap', colors)

    colors1 = plt.cm.nipy_spectral_r(np.linspace(0, 1, 255))
    colors2 = plt.cm.Set1_r(np.linspace(0, 1, 1))
    colors = np.vstack((colors2, colors1))
    cmap1 = mcolors.LinearSegmentedColormap.from_list('my_colormap', colors)

    sns.set_style('darkgrid')
    sns.set_context("notebook")

    plt.close(0)
    fig, ax = plt.subplots(num = 0, figsize = (12,10),dpi=args['dpi'],
                           nrows = 3, ncols = 2)

    ################################################
    #           SWI                                #
    ################################################
    mymap = copy.deepcopy(cmap)
    pmask = masks[plotorder[0]]['mask']
    ixo = pmask == 0
    accum[ixo] = np.nan
    mymap.set_bad('white',1.)

    r = ~np.isnan(accum)
    r[r] &= accum[r] < 0.001
    accum[r] = 0

    h = ax[0,0].imshow(accum,  cmap = mymap, clim=clims)

    for name in masks:
        ax[0,0].contour(masks[name]['mask'],cmap = 'Greys',linewidths = 1)

    h.axes.get_xaxis().set_ticks([])
    h.axes.get_yaxis().set_ticks([])
    divider = make_axes_locatable(ax[0,0])
    cax = divider.append_axes("right", size="4%", pad=0.2)
    cbar = plt.colorbar(h, cax = cax)
    cbar.set_label('[{}]'.format(args['depthlbl']))
    h.axes.set_title('Accumulated SWI')
    xlims = args['xlims']
    if len(plotorder) == 1:
        sumorder = plotorder
        swid = 0.4
        wid = [-0.1, 0.1]
    elif len(plotorder) <= 4:
        sumorder = plotorder[1::]
        swid = 0.25
        wid = np.linspace(-0.3,0.3,len(sumorder))
    elif len(plotorder) == 5:
        sumorder = plotorder[1::]
        swid = 0.2
        wid = np.linspace(-0.3,0.3,len(sumorder))
    elif len(plotorder) > 5:
        sumorder = plotorder[1::]
        swid = 0.1
        wid = np.linspace(-0.4,0.4,len(sumorder))

    for iters,name in enumerate(sumorder):
        lbl = labels[name]
        if len(plotorder) == 1:
            ax[0,1].bar(range(0,len(edges)),
                    accum_byelev[name],
                    color = barcolors[iters], width = swid, edgecolor = 'k', label = lbl)
        else:
            ax[0,1].bar(range(0,len(edges))-wid[iters],
                    accum_byelev[name],
                    color = barcolors[iters], width = swid, edgecolor = 'k', label = lbl)

    ax[0,1].xaxis.set_ticks(range(0,len(edges)))
    ax[0,1].set_xlim((xlims[0]-0.5,xlims[1]-0.5))

    edges_lbl = []
    for i in range(0,len(edges)):
        edges_lbl.append(str(int(edges[int(i)])))

    ax[0,1].set_xticklabels(str(i) for i in edges_lbl)
    for tick in ax[0,1].get_xticklabels():
        tick.set_rotation(30)

    ax[0,1].set_xlim((xlims[0]-0.5,xlims[1]-0.5))
    ax[0,1].set_ylabel('[{}]'.format(args['depthlbl']))
    ax[0,1].yaxis.set_label_position("right")
    ax[0,1].yaxis.tick_right()
    ax[0,1].set_ylim((0,yMax))

    if sum(sum(r)) > 1000:
        patches = [mpatches.Patch(color='grey', label='no SWI')]
        ax[0,0].legend(handles=patches, bbox_to_anchor=(lims.pbbx, 0.05),
                  loc=2, borderaxespad=0. )

    if len(plotorder) > 1:
        ax[0,1].legend(loc=(lims.legx,lims.legy2),markerscale = 0.5)

    ################################################
    #           Precip                             #
    ################################################

    mymap1 = copy.deepcopy(cmap1)
    pmask = masks[plotorder[0]]['mask']
    ixo = pmask == 0
    precip[ixo] = np.nan
    mymap1.set_bad('white',1.)
    r = ~np.isnan(precip)
    r[r] &= precip[r] < 0.001
    precip[r] = 0

    h2 = ax[1,0].imshow(precip, interpolation='none', cmap = mymap1, clim = clims)

    for name in masks:
        ax[1,0].contour(masks[name]['mask'],cmap = "Greys",linewidths = 1)

    h2.axes.get_xaxis().set_ticks([])
    h2.axes.get_yaxis().set_ticks([])
    divider = make_axes_locatable(ax[1,0])
    cax = divider.append_axes("right", size="5%", pad=0.2)
    cbar = plt.colorbar(h2, cax = cax)
    cbar.set_label(r'[{}]'.format(args['depthlbl']))

    h2.axes.set_title('Precipitation')

    for iters,name in enumerate(sumorder):
        lbl = name

        if len(plotorder) == 1:
            ax[1,1].bar(range(0,len(edges)),
                    precip_byelev[name],
                    color = barcolors[iters], width = swid, edgecolor = 'k', label = lbl)
        else:
            ax[1,1].bar(range(0,len(edges))-wid[iters],
                    precip_byelev[name],
                    color = barcolors[iters], width = swid, edgecolor = 'k', label = lbl)

        ax[1,1].set_xlim((xlims[0]-0.5,xlims[1]-0.5))

    ax[1,1].xaxis.set_ticks(range(0,len(edges)))
    ax[1,1].set_xlim((xlims[0]-0.5,xlims[1]-0.5))

    edges_lbl = []
    for i in range(0,len(edges)):
        edges_lbl.append(str(int(edges[int(i)])))

    ax[1,1].set_xticklabels(str(i) for i in edges_lbl)
    for tick in ax[1,1].get_xticklabels():
        tick.set_rotation(30)

    ax[1,1].set_xlim((xlims[0]-0.5,xlims[1]-0.5))
    ax[1,1].set_ylim((0,yMax))
    ax[1,1].set_ylabel(r'[{}]'.format(args['depthlbl']))
    ax[1,1].yaxis.set_label_position("right")
    ax[1,1].tick_params(axis='x')
    ax[1,1].tick_params(axis='y')
    ax[1,1].yaxis.tick_right()

    patches = [mpatches.Patch(color='grey', label='no precip')]

    ################################################
    #           rain                                #
    ################################################

    ix = rain_byelev < 0
    rain_byelev[ix] = 0

    mymap = copy.deepcopy(cmap1)
    pmask = masks[plotorder[0]]['mask']
    ixo = pmask == 0
    rain[ixo] = np.nan
    mymap.set_bad('white',1.)
    r = ~np.isnan(rain)
    r[r] &= rain[r] < 0.001
    rain[r] = 0

    h2 = ax[2,0].imshow(rain, interpolation='none', cmap = mymap, clim = clims)

    for name in masks:
        ax[2,0].contour(masks[name]['mask'],cmap = "Greys",linewidths = 1)

    h2.axes.get_xaxis().set_ticks([])
    h2.axes.get_yaxis().set_ticks([])
    divider = make_axes_locatable(ax[2,0])
    cax = divider.append_axes("right", size="5%", pad=0.2)
    cbar = plt.colorbar(h2, cax = cax)
    cbar.set_label(r'[{}]'.format(args['depthlbl']))

    h2.axes.set_title('Rain')

    for iters,name in enumerate(sumorder):
        lbl = name
        if len(plotorder) == 1:
            ax[2,1].bar(range(0,len(edges)),
                    rain_byelev[name],
                    color = barcolors[iters], width = swid, edgecolor = 'k', label = lbl)
        else:
            ax[2,1].bar(range(0,len(edges))-wid[iters],
                    rain_byelev[name],
                    color = barcolors[iters], width = swid, edgecolor = 'k', label = lbl)

        ax[2,1].set_xlim((xlims[0]-0.5,xlims[1]-0.5))

    ax[2,1].xaxis.set_ticks(range(0,len(edges)))
    ax[2,1].set_xlim((xlims[0]-0.5,xlims[1]-0.5))

    edges_lbl = []
    for i in range(0,len(edges)):
        edges_lbl.append(str(int(edges[int(i)])))

    ax[2,1].set_xticklabels(str(i) for i in edges_lbl)
    for tick in ax[2,1].get_xticklabels():
        tick.set_rotation(30)

    ax[2,1].set_xlim((xlims[0]-0.5,xlims[1]-0.5))
    ax[2,1].set_ylim((0,yMax))
    ax[2,1].set_ylabel(r'[{}]'.format(args['depthlbl']))
    ax[2,1].set_xlabel('elevation [{}]'.format(args['elevlbl']))
    ax[2,1].yaxis.set_label_position("right")
    ax[2,1].tick_params(axis='x')
    ax[2,1].tick_params(axis='y')
    ax[2,1].yaxis.tick_right()

    plt.suptitle(args['title'])
    fig.subplots_adjust(top=0.92)

    fig_name_short = 'precip_depth_'
    fig_name = '{}{}{}.png'.format(args['figs_path'],fig_name_short,args['directory'])
    if logger is not None:
        logger.info(' saving {}'.format(fig_name))
    snowav.framework.figures.save_fig(fig, fig_name)

    return fig_name_short



import os
__version__ = '0.11.15'
__core_config__ = os.path.abspath(os.path.dirname(__file__) + '/config/CoreConfig.ini')
__recipes__ = os.path.abspath(os.path.dirname(__file__) + '/config/recipes.ini')
__config_titles__ = {'snowav':'Overview',
                     'database':'Database connection',
                     'run':'Process directory and report period',
                     'validate':'Snow pillow validation figure',
                     'diagnostics':'Additional model inputs and outputs figure',
                     'plots':'Figures',
                     'inflow':'Reservoir inflow and SWI comparison',
                     'report':'Report',
                     'forecast':'WRF forecast',
                     'query':'Query existing database records'}

from . import framework
from . import database
from . import plotting
from . import report
from . import utils

__config_header__ = utils.utilities.get_config_header()




import seaborn as sns
from matplotlib import pyplot as plt
import snowav.framework.figures
import matplotlib.dates as mdates
from datetime import datetime

def inputs(args, logger):
    '''
    Simple summary of smrf inputs.

    Args
    ----------
    args : dict
        dictionary with required inputs, see swi() figure for more information.

    logger : list
        snowav logger

    '''

    dfmt = mdates.DateFormatter('%b-%d')
    sns.set_style('whitegrid')

    values = args['inputs']
    barcolors = args['barcolors']
    lines = ['-',':','--','-.','-',':','--','-.','-',':','--','-.']

    plt.close(0)
    f, a = plt.subplots(num=0,figsize=(8,6),dpi=args['dpi'],nrows=3,ncols=3)
    a = a.flatten()

    p = []
    for v in values[args['var_list'][0]][args['inputs_basins'][0]]:
        if 'percentile' in v:
            p.append(v)

    for idx, val in enumerate(args['var_list']):

        for ixn, name in enumerate(args['inputs_basins']):

            for ixs, stat in enumerate(args['inputs_methods']):

                clr = barcolors[ixn]
                lstyle = lines[ixs]
                vlstyle = lines[ixs + 1]

                lbl = '__nolabel__'

                if idx == 0 and stat == args['inputs_methods'][0]:
                    lbl = name

                if idx == 1 and name == args['inputs_basins'][0]:
                    lbl = stat

                if 'percentile' not in stat:
                    if idx == 0:
                        a[idx].plot(values[val][name][stat]['value'],
                                    label = lbl,
                                    color = clr,
                                    linestyle = lstyle,
                                    linewidth = 0.5)

                    else:
                        a[idx].plot(values[val][name][stat]['value'],
                                    label = lbl,
                                    color = clr,
                                    linestyle = lstyle,
                                    linewidth = 0.5)

                else:
                    x = values[val][name][stat]['value'].index
                    y1 = values[val][name][p[0]]['value']
                    y2 = values[val][name][p[1]]['value']

                    if not y1.empty or not y2.empty:
                        a[idx].fill_between(x, y1, y2, color=clr, alpha='0.075')

                a[idx].set_xlim((datetime(args['wy']-1, 10, 1), args['end_date']))
                a[idx].set_title(val, fontsize = 6)
                a[idx].xaxis.set_major_formatter(dfmt)
                for tick in a[idx].get_xticklabels():
                    tick.set_rotation(45)
                    tick.set_fontsize(6)

                for tick in a[idx].get_yticklabels():
                    tick.set_fontsize(6)

    for n in range(idx+1,len(a)):
        f.delaxes(a[n])

    a[0].legend(loc=3,fontsize=6)
    a[1].legend(loc=3,fontsize=6)

    plt.tight_layout()

    fig_name_short = 'inputs_'
    fig_name = '{}{}{}.png'.format(args['figs_path'],fig_name_short,
                                   args['directory'])
    if logger is not None:
        logger.info(' saving {}'.format(fig_name))
    snowav.framework.figures.save_fig(f, fig_name)

    for idx, val in enumerate(values.keys()):
        a[idx].set_xlim((args['start_date'], args['end_date']))

    fig_name_short = 'inputs_period_'
    fig_name = '{}{}{}.png'.format(args['figs_path'],fig_name_short,
                                   args['directory'])
    if logger is not None:
        logger.info(' saving {}'.format(fig_name))
    snowav.framework.figures.save_fig(f, fig_name)




from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
from matplotlib import pyplot as plt
import matplotlib.colors as mcolors
import seaborn as sns
import copy
import cmocean
import matplotlib.patches as mpatches
import pandas as pd
import snowav.framework.figures

def density(args, logger = None):
    '''
    Density figure.

    Args
    ----------
    args : dict
        dictionary with required inputs, see swi() figure for more information.

    logger : dict

    '''

    if args['print']:
        print('density() figure args:\n','omitting masks, image, and density...\n')
        for name in args.keys():
            if name not in ['masks','image', 'density']:
                print(name, ': ', args[name])

    masks = args['masks']
    image = args['image']
    density = args['density']
    plotorder = args['plotorder']
    lims = args['lims']
    edges = args['edges']
    labels = args['labels']
    barcolors = args['barcolors']

    qMin,qMax = np.nanpercentile(image,[5,95])
    clims = (qMin,qMax)

    sns.set_style('darkgrid')
    sns.set_context("notebook")

    cvalue = copy.deepcopy(image)

    mymap = plt.cm.get_cmap('BuGn', 8)

    pmask = masks[plotorder[0]]['mask']
    ixo = pmask == 0
    cvalue[ixo] = np.nan
    mymap.set_bad('white',1.)

    ixf = cvalue == 0
    cvalue[ixf] = -1
    mymap.set_under('lightgrey',1.)

    plt.close(4)
    fig,(ax,ax1) = plt.subplots(num=4, figsize = args['figsize'],
                                dpi=args['dpi'], nrows = 1, ncols = 2)
    h = ax.imshow(cvalue, vmin = 200, vmax = 500, cmap = mymap)

    for name in masks:
        ax.contour(masks[name]['mask'],cmap = 'Greys',linewidths = 1)

    h.axes.get_xaxis().set_ticks([])
    h.axes.get_yaxis().set_ticks([])
    divider = make_axes_locatable(ax)

    cax = divider.append_axes("right", size="4%", pad=0.2)
    cbar = plt.colorbar(h, cax = cax)
    cbar.ax.tick_params()
    cbar.set_label('[kg/$m^3$]')

    h.axes.set_title(args['title'])

    sep = 0.05
    wid = 1/len(plotorder)-sep
    widths = np.arange((-1 + wid), (1 - wid), wid)

    for i,name in enumerate(lims.sumorder):
        for iter,edge in enumerate(edges):
            data = density[name][edge].flatten()
            data = data[~np.isnan(data)]

            if np.nansum(data) > 0:
                bp = ax1.boxplot(data,
                                 positions=[iter + widths[i]],
                                 widths=wid)

                for element in ['boxes', 'whiskers', 'caps']:
                    plt.setp(bp[element], color=barcolors[i])

    ax1.set_xticks(np.arange(0,len(edges)))
    ax1.set_xticklabels([str(x) for x in edges])
    ax1.set_xlim(args['xlims'][0],args['xlims'][1])

    for tick in ax1.get_xticklabels():
        tick.set_rotation(30)

    ax1.set_title('Elevation distribution')
    ax1.set_ylabel(r'density [kg/$m^3$]')

    ax1.yaxis.set_label_position("right")
    ax1.yaxis.tick_right()

    plt.tight_layout()
    fig.subplots_adjust(top=0.88)

    if sum(sum(ixf)) > 1000:
        patches = [mpatches.Patch(color='grey', label='snow free')]
        ax.legend(handles=patches, bbox_to_anchor=(lims.pbbx, 0.05),
                  loc=2, borderaxespad=0. )

    fig_name_short = 'density_'
    fig_name = '{}{}{}.png'.format(args['figs_path'],fig_name_short,args['directory'])
    if logger is not None:
        logger.info(' saving {}'.format(fig_name))

    snowav.framework.figures.save_fig(fig, fig_name)

    return fig_name_short

    ###########################################
    #   2nd density figure
    ###########################################
    #
    # plt.close(4)
    # f = plt.figure(num=4, figsize = snow.figsize, dpi=snow.dpi)
    # a = plt.gca()
    #
    # for iters,name in enumerate(snow.plotorder):
    #     if iters == 0:
    #         lname = '8k ft'
    #     else:
    #         lname = '__nolabel__'
    #
    #     a.plot(density_summary[name],
    #            color = snow.barcolors[iters],
    #            label = name)
    #     a.plot(density_summary_8[name],
    #            color = snow.barcolors[iters],
    #            linewidth = 0.75,
    #            linestyle = ':',
    #            label = lname)
    #
    # if snow.flight_dates is not None:
    #     for d in snow.flight_dates:
    #         a.axvline(x=d,linestyle = ':',linewidth = 0.75, color = 'k')
    #
    # a.legend(loc=2)
    # a.set_ylabel('density [kg/$m^3$]')
    # a.set_xlim((snow.start_date, snow.end_date))
    #
    # for tick in a.get_xticklabels():
    #     tick.set_rotation(30)
    #
    # f.tight_layout()
    # f.savefig('{}density_change_{}.png'.format(snow.figs_path,snow.name_append))



import cmocean
import numpy as np
from matplotlib import pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
from mpl_toolkits.axes_grid1 import make_axes_locatable
import os
import pandas as pd
import seaborn as sns

import snowav.framework.figures


def swe_volume(args, logger=None):
    """
    Current SWE depth and volume.

    Note: scripts/sample_figure.py and command line tool call this
    function. If changes are made here update those calls as necessary.

    Args
    ------
    args : dict
        dictionary with required inputs, see swi() figure for more information.

    logger : dict
        snowav logger

    """

    # print to screen for each figure if desired
    if args['print']:
        print('swe_volume() figure args:\n', 'omitting masks and image...\n')
        for name in args.keys():
            if name not in ['masks', 'image']:
                print(name, ': ', args[name])

    image = args['image']
    masks = args['masks']
    df = args['df']
    plotorder = args['plotorder']
    lims = args['lims']
    edges = args['edges']
    labels = args['labels']
    barcolors = args['barcolors']

    q_min, q_max = np.nanpercentile(image, [0, args['percent_max']])
    clims = (q_min, q_max)
    pmask = masks[plotorder[0]]['mask']
    ixo = pmask == 0
    ixz = image == 0
    image[ixz] = -1

    colorsbad = plt.cm.Set2_r(np.linspace(0., 1, 1))
    colors1 = cmocean.cm.haline_r(np.linspace(0., 1, 254))
    colors = np.vstack((colorsbad, colors1))
    mymap = mcolors.LinearSegmentedColormap.from_list('my_colormap', colors)

    image[ixo] = np.nan
    mymap.set_bad('white', 1.)

    sns.set_style('darkgrid')
    sns.set_context("notebook")

    plt.close(21)
    fig, (ax, ax1) = plt.subplots(num=21, figsize=args['figsize'],
                                  facecolor='white', dpi=args['dpi'],
                                  nrows=1, ncols=2)

    h = ax.imshow(image, clim=clims, cmap=mymap)

    for name in masks:
        ax.contour(masks[name]['mask'], cmap="Greys", linewidths=1)

    h.axes.get_xaxis().set_ticks([])
    h.axes.get_yaxis().set_ticks([])
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.2)
    cbar = plt.colorbar(h, cax=cax)
    cbar.set_label('[{}]'.format(args['depthlbl']))
    h.axes.set_title(args['title'])

    patches = [mpatches.Patch(color='grey', label='snow free')]

    # If there is meaningful snow-free area, include path and label
    if sum(sum(ixz)) > 1000:
        patches = [mpatches.Patch(color='grey', label='snow free')]
        ax.legend(handles=patches, bbox_to_anchor=(lims.pbbx, 0.05),
                  loc=2, borderaxespad=0.)

    ax1.legend(loc=(lims.legx, lims.legy), markerscale=0.5)

    for iters, name in enumerate(lims.sumorder):
        if args['dplcs'] == 0:
            ukaf = str(np.int(np.nansum(df[name].values)))
        else:
            ukaf = str(np.round(np.nansum(df[name].values), args['dplcs']))

        if iters == 0:

            ax1.bar(range(0, len(edges)), df[name],
                    color=barcolors[iters],
                    edgecolor='k',
                    label=labels[name] + ': {} {}'.format(ukaf, args['vollbl']))

        else:
            ax1.bar(range(0, len(edges)), df[name],
                    bottom=pd.DataFrame(df[lims.sumorder[0:iters]]).sum(axis=1).values,
                    color=barcolors[iters],
                    edgecolor='k',
                    label=labels[name] + ': {} {}'.format(ukaf, args['vollbl']))

    ylims = ax1.get_ylim()
    ymax = ylims[1] + ylims[1] * 0.5
    ymin = 0
    ax1.set_ylim((ymin, ymax))

    # Keep these to use in cold_content()
    swe_ylims = (ymin, ymax)

    ax1.xaxis.set_ticks(range(0, len(edges)))
    plt.tight_layout()
    ax1.set_xlim((args['xlims'][0] - 0.5, args['xlims'][1] - 0.5))

    edges_lbl = []
    for i in range(0, len(edges)):
        edges_lbl.append(str(int(edges[int(i)])))

    ax1.set_xticklabels(str(i) for i in edges_lbl)
    for tick in ax1.get_xticklabels():
        tick.set_rotation(30)

    ax1.set_xlim((args['xlims'][0] - 0.5, args['xlims'][1] - 0.5))
    ax1.set_xlabel('elevation [{}]'.format(args['elevlbl']))
    ax1.set_ylabel('{}'.format(args['vollbl']))
    ax1.yaxis.set_label_position("right")
    ax1.yaxis.tick_right()
    ax1.legend(loc=2, fontsize=10)

    for tick in ax1.get_xticklabels():
        tick.set_rotation(30)

    fig.tight_layout()
    fig.subplots_adjust(top=0.92, wspace=0.2)

    fig_name_short = 'swe_volume_'
    fig_name = os.path.join(args['figs_path'],
                            '{}{}.png'.format(fig_name_short,
                                              args['directory']))
    if logger is not None:
        logger.info(' saving {}'.format(fig_name))

    snowav.framework.figures.save_fig(fig, fig_name)

    if 'show' in args.keys() and args['show']:
        plt.show()

    return fig_name_short, swe_ylims




import os
from datetime import datetime
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns
import copy
import matplotlib.dates as mdates
import snowav.framework.figures

def compare_runs(args, logger = None):
    '''
    Basin total values for existing database runs.

    Currently only plotting SWE and SWI.

    Args
    ----------
    args : dict
        dictionary with required inputs, see swi() figure for more information.

    logger : list
        snowav logger

    '''

    start_date = datetime(args['wy'] - 1,10,1)
    end_date = args['end_date']
    bid = args['plotorder'][0]

    sns.set_style('darkgrid')
    sns.set_context("notebook")

    for i, var in enumerate(args['variables']):

        plt.close(i)
        fig = plt.figure(num = i, figsize = args['figsize'], dpi = args['dpi'])
        ax = plt.gca()

        formatter = mdates.DateFormatter('%b')
        ax.xaxis.set_major_formatter(formatter)

        if 'z' in var:
            lbl = 'mm'
        if ('vol' in var) or 'avail' in var:
            lbl = args['vollbl']
        if var == 'density':
            lbl = 'kg/m^3'

        if var == 'swe_vol':
            title = 'Basin Total SWE'
        if var == 'swi_vol':
            title = 'Basin Total SWI'

        for name in args['dict'][var]:
            ax.plot(args['dict'][var][name], label = name)

        if args['flag']:
            for ix,d in enumerate(args['flight_dates']):
                if ix == 0:
                    lb = 'wy{} flight update'.format(args['wy'])
                else:
                    lb = '__nolabel__'

                ax.axvline(x = d, linestyle = ':', linewidth = 0.75,
                           color = 'k', label = lb)

        for tick in ax.get_xticklabels():
            tick.set_rotation(30)

        formatter = mdates.DateFormatter('%b')
        ax.xaxis.set_major_formatter(formatter)
        ax.legend()
        ax.set_ylabel('[{}]'.format(lbl))
        ax.set_xlim((datetime(args['wy']-1,10,1),datetime(args['wy'],8,1)))

        ax.set_title(title)
        plt.tight_layout()

        fig_name_short = 'compare_{}_'.format(var)
        fig_name = '{}{}{}.png'.format(args['figs_path'],fig_name_short,args['directory'])
        if logger is not None:
            logger.info(' saving {}'.format(fig_name))
        snowav.framework.figures.save_fig(fig, fig_name)

    return fig_name_short



from datetime import datetime
from jinja2 import FileSystemLoader
from latex.jinja2 import make_env
from latex import build_pdf
import os

import snowav
from snowav.database.database import collect


def report(cfg):
    """ Create the pdf report.

    The latex formats for figures, tables, and summary paragraphs are created
    by templates in snowav/report/. Data is pulled from the snowav database.

    To add a figure to the report:
    - create the plot (of course)
    - add to variables dict in this file, with the same naming convention
        variables['PDEP_FIG'] = 'precip_depth%s.png'%(cfg.directory)
    - add figure template to snowav_report.tex
        \VAR{PDEP_FIG_TPL}
    - add to section_dict
    - add template to snowav/report/figs/
    - add to snowav/config/CoreConfig.py [report] exclude_figs
    - if the figure may not exist (such as flt_diff, or those that require
      process() to run), address that with some form of exception before
      creating the pdf

    Args
    ------
    cfg {class}: config class
    """

    # bid = cfg.basins[cfg.plotorder[0]]['basin_id']
    basins = cfg.basins
    wy_start = datetime(cfg.wy-1, 10, 1)
    start_date = cfg.start_date
    end_date = cfg.end_date
    run_name = cfg.run_name
    # edges = cfg.edges
    plotorder = cfg.plotorder
    dpts = int(cfg.dplcs)
    ddpts = int(cfg.rep_dplcs)
    cnx = cfg.connector

    # variables to pass to latex file
    variables = {}

    dbval = collect(cnx, plotorder[0], basins, wy_start, end_date, 'swi_vol',
                    run_name, 'total', 'sum')
    variables['TOTAL_SWI'] = dbval.sum().values.round(decimals=dpts)[0]

    dbval = collect(cnx, plotorder[0], basins, start_date, end_date, 'swi_vol',
                    run_name, 'total', 'sum')
    variables['PER_SWI'] = dbval.sum().values.round(decimals=dpts)[0]

    dbval = collect(cnx, plotorder[0], basins, start_date, end_date, 'swe_vol',
                    run_name, 'total', 'end')
    variables['TOTAL_SWE'] = dbval.sum().values.round(decimals=dpts)[0]

    dbval = collect(cnx, plotorder[0], basins, start_date, end_date, 'swe_avail',
                    run_name, 'total', 'end')
    variables['TOTAL_SAV'] = dbval.sum().values.round(decimals=dpts)[0]

    dbval = collect(cnx, plotorder[0], basins, start_date, end_date, 'swe_z',
                    run_name, 'total', 'end')
    variables['TOTAL_PM'] = dbval.sum().values.round(decimals=ddpts)[0]

    dbval = collect(cnx, plotorder[0], basins, wy_start, end_date, 'precip_z',
                    run_name, 'total', 'sum')
    variables['TOTALPRE_PM'] = dbval.sum().values.round(decimals=ddpts)[0]

    s = collect(cnx, plotorder[0], basins, start_date, start_date, 'swe_vol',
                run_name, 'total', 'end')
    e = collect(cnx, plotorder[0], basins, start_date, end_date, 'swe_vol',
                run_name, 'total', 'end')
    start_swe = s.sum().values.round(decimals=dpts)[0]
    end_swe = e.sum().values.round(decimals=dpts)[0]
    diff = end_swe - start_swe
    variables['TOTAL_SDEL'] = diff
    variables['TOTAL_PVOL'] = '100'

    if float(end_swe) - float(start_swe) > 0:
        variables['SIGN'] = r'$+$'
    if float(end_swe) - float(start_swe) == 0.0:
        variables['SIGN'] = ''
    if float(end_swe) - float(start_swe) < 0.0:
        variables['SIGN'] = r'-'

    dbval = collect(cnx, plotorder[0], basins, wy_start, end_date, 'rain_z',
                    run_name, 'total', 'sum')
    total_rain = dbval.sum().values.round(decimals=ddpts)[0]

    if variables['TOTALPRE_PM'] != 0:
        variables['TOTAL_RAT'] = str(int((total_rain / variables['TOTALPRE_PM']) * 100))
    else:
        variables['TOTAL_RAT'] = '-'
        variables['TOTALPRE_PM'] = '-'

    report_time = datetime.now().strftime("%Y-%-m-%-d %H:%M")

    numsubs = range(1, len(cfg.plotorder))

    for n, sub in zip(numsubs, cfg.plotorder[1:]):
        swiind = 'SUB' + str(n) + '_SWI'
        perswiind = 'SUB' + str(n) + '_PERSWI'
        sweind = 'SUB' + str(n) + '_SWE'
        avsweind = 'SUB' + str(n) + '_SAV'
        swedel = 'SUB' + str(n) + '_SDEL'
        pm = 'SUB' + str(n) + '_PM'
        prepm = 'SUB' + str(n) + 'PRE_PM'
        rain = 'SUB' + str(n) + 'RAI'
        ratio = 'SUB' + str(n) + '_RAT'
        pvol = 'SUB' + str(n) + '_PVOL'

        dbval = collect(cnx, sub, basins, wy_start, end_date, 'swi_vol',
                        run_name, 'total', 'sum')
        variables[swiind] = dbval.sum().values.round(decimals=dpts)[0]

        dbval = collect(cnx, sub, basins, start_date, end_date, 'swi_vol',
                        run_name, 'total', 'sum')
        variables[perswiind] = dbval.sum().values.round(decimals=dpts)[0]

        dbval = collect(cnx, sub, basins, start_date, end_date, 'swe_vol',
                        run_name, 'total', 'end')
        variables[sweind] = dbval.sum().values.round(decimals=dpts)[0]

        dbval = collect(cnx, sub, basins, start_date, end_date, 'swe_avail',
                        run_name, 'total', 'end')
        variables[avsweind] = dbval.sum().values.round(decimals=dpts)[0]

        dbval = collect(cnx, sub, basins, start_date, start_date, 'swe_vol',
                        run_name, 'total', 'end')
        start_swe = dbval.sum().values.round(decimals=dpts)[0]
        dbval = collect(cnx, sub, basins, start_date, end_date, 'swe_vol',
                        run_name, 'total', 'end')
        end_swe = dbval.sum().values.round(decimals=dpts)[0]
        variables[swedel] = end_swe - start_swe

        dbval = collect(cnx, sub, basins, start_date, end_date, 'swe_z',
                        run_name, 'total', 'end')
        variables[pm] = dbval.sum().values.round(decimals=ddpts)[0]

        dbval = collect(cnx, sub, basins, wy_start, end_date, 'precip_z',
                        run_name, 'total', 'sum')
        variables[prepm] = dbval.sum().values.round(decimals=ddpts)[0]

        dbval = collect(cnx, sub, basins, wy_start, end_date, 'rain_z',
                        run_name, 'total', 'sum')
        variables[rain] = dbval.sum().values.round(decimals=ddpts)[0]

        if (end_swe > 0) and (variables['TOTAL_SWE'] > 0):
            variables[pvol] = end_swe / variables['TOTAL_SWE'] * 100
        else:
            variables[pvol] = '-'

        if variables[prepm] != 0.0:
            variables[ratio] = str(int((variables[rain] / variables[prepm]) * 100))
        else:
            variables[ratio] = '0'

    # Upper case variables are used in the LaTex file,
    # lower case versions are assigned here

    # untested - if report title contains comma?
    if isinstance(cfg.rep_title, list):
        title = cfg.rep_title[0]
        for s in cfg.rep_title[1::]:
            title = title + ', ' + s
        variables['REPORT_TITLE'] = title

    else:
        if cfg.flt_flag and cfg.flight_figs:
            fst = ' Model snow depths were updated with ASO snow depths on '
            tst = ''
            for i, d in enumerate(cfg.flight_diff_dates):
                if d <= cfg.end_date:
                    dn = d.date().strftime("%m/%d")
                    if len(cfg.flight_diff_dates) == 1:
                        fst = fst + dn + '.'
                        tst += dn

                    if ((len(cfg.flight_diff_dates) > 1) and
                            (i < len(cfg.flight_diff_dates) - 1)):
                        fst = fst + dn + ', '
                        tst += dn + ', '

                    if ((len(cfg.flight_diff_dates) > 1) and
                            (i == len(cfg.flight_diff_dates) - 1)):
                        fst = fst + 'and ' + dn + '.'
                        tst += dn

                else:
                    fst = fst.split(dn)[0] + 'and ' + dn + '.'
                    break

            variables['REPORT_TITLE'] = (cfg.rep_title +
                                         r' \\ ASO Updates {}'.format(tst))
            variables['FLTSENT'] = fst

        else:
            variables['REPORT_TITLE'] = cfg.rep_title
            variables['FLTSENT'] = ''

    variables['REPORT_TIME'] = report_time
    variables['WATERYEAR'] = str(cfg.wy)
    variables['UNITS'] = cfg.vollbl
    variables['VOLLBL'] = cfg.vollbl
    variables['DEPLBL'] = cfg.depthlbl
    variables['START_DATE'] = cfg.report_start.date().strftime("%B %-d")
    variables['END_DATE'] = cfg.report_date.date().strftime("%B %-d")
    variables['SWE_IN'] = variables['TOTAL_PM']
    variables['SWI_IN'] = variables['TOTAL_SWI']
    variables['FIG_PATH'] = cfg.figs_path
    variables['SWI_FIG'] = 'swi_{}.png'.format(cfg.directory)
    variables['CHANGES_FIG'] = 'swe_change_{}.png'.format(cfg.directory)
    variables['TOTALS_FIG'] = 'basin_total_{}.png'.format(cfg.directory)
    variables['MULTITOTSWE_FIG'] = 'compare_swe_vol_{}.png'.format(cfg.directory)
    variables['DENSITY_FIG'] = 'density_{}.png'.format(cfg.directory)
    variables['PDEP_FIG'] = 'precip_depth_{}.png'.format(cfg.directory)
    variables['VALID_FIG'] = cfg.stn_validate_fig_name
    variables['COLD_FIG'] = 'cold_content_{}.png'.format(cfg.directory)
    variables['SWE_FIG'] = 'swe_volume_{}.png'.format(cfg.directory)
    variables['VERSION'] = snowav.__version__

    if (cfg.update_file is not None) and cfg.flt_flag and cfg.flight_figs:
        for name in cfg.flight_diff_fig_names:
            variables['DFLT_FIG'] = name

    if cfg.forecast_flag:
        variables['FORE_START_DATE'] = cfg.for_start_date.date().strftime("%B %-d")
        variables['FORE_DATE'] = cfg.for_end_date.date().strftime("%B %-d")
        variables['SWEFORECAST_FIG'] = 'swe_volume_{}.png'.format(cfg.directory + '_forecast')
        variables['SWIFORECAST_FIG'] = 'swi_{}.png'.format(cfg.directory + '_forecast')
        variables['CHANGESFORECAST_FIG'] = 'swe_change_{}.png'.format(cfg.directory + '_forecast')
        variables['TOTALSFORECAST_FIG'] = 'basin_total_{}.png'.format(cfg.directory + '_forecast')
        variables['PDEPFORECAST_FIG'] = 'precip_depth_{}.png'.format(cfg.directory + '_forecast')

    if cfg.report_diagnostics:
        variables['DIAGNOSTICS_FIG'] = 'diagnostics_{}'.format(cfg.directory)
        variables['INPUTS_FIG'] = 'inputs_period_{}'.format(cfg.directory)

    if cfg.subs_fig is not None:
        variables['SUBBASINS_FIG'] = '{}'.format(cfg.subs_fig)

    # Logos
    variables['ARSLOGO'] = os.path.join(cfg.figs_tpl_path, 'ARS.jpg')
    variables['ASOLOGO'] = os.path.join(cfg.figs_tpl_path, 'ASO.jpg')
    variables['USDALOGO'] = os.path.join(cfg.figs_tpl_path, 'USDA.png')
    variables['JPLLOGO'] = os.path.join(cfg.figs_tpl_path, 'JPL.jpg')
    variables['CDWRLOGO'] = os.path.join(cfg.figs_tpl_path, 'CDWR.png')
    variables['USBRLOGO'] = os.path.join(cfg.figs_tpl_path, 'USBR.jpg')
    variables['NRCSLOGO'] = os.path.join(cfg.figs_tpl_path, 'NRCS.jpg')
    variables['KRWALOGO'] = os.path.join(cfg.figs_tpl_path, 'KRWA.jpg')
    variables['FRIANTLOGO'] = os.path.join(cfg.figs_tpl_path, 'FRIANT.jpg')
    variables['AWSMLOGO'] = os.path.join(cfg.figs_tpl_path, 'logo.png')

    dfind = [str(i) for i in cfg.edges]
    dfindt = [str(i) for i in cfg.edges] + ['total']
    colstr = 'l' + 'r' * len(cfg.plotorder)

    if len(cfg.plotorder) > 5:
        spacecmd = r'\resizebox{\textwidth}{!}{'
    else:
        spacecmd = r'{'

    # ntables = len(cfg.tables)
    mtables = 2
    ptables = 0

    if 'swe_depth' in cfg.tables:

        dbval = collect(cnx, plotorder, basins, start_date, end_date, 'swe_z', run_name, dfindt, 'end')
        dbval = dbval.rename(columns=cfg.labels)
        swe_byelev = dbval.round(decimals=dpts)
        swe_byelev.rename(index={'total': 'mean'}, inplace=True)
        swe_byelev.index.name = 'Elevation'
        variables['SWE_BYELEV'] = (r' \normalsize \textbf{SWE [%s], %s}\\ \vspace{0.1cm} \\'
                                   % (cfg.depthlbl, cfg.report_date.date().strftime("%Y-%-m-%-d")) +
                                   spacecmd + swe_byelev.to_latex(na_rep='-', column_format=colstr) +
                                   r'} \\ \footnotesize{\textbf{Table %s:} SWE depth.}' % (str(mtables)))
        mtables += 1
        ptables += 1

    else:
        variables['SWE_BYELEV'] = ''

    if 'swe_vol' in cfg.tables:
        ptables += 1

        if ptables == 2:
            clrpage = r'\clearpage'
        else:
            clrpage = ''
        dbval = collect(cnx, plotorder, basins, start_date, end_date, 'swe_vol', run_name, dfindt, 'end')
        dbval = dbval.rename(columns=cfg.labels)
        swe_byelev = dbval.round(decimals=dpts)
        swe_byelev.index.name = 'Elevation'
        variables['SWEVOL_BYELEV'] = (r' \normalsize \textbf{SWE [%s], %s}\\ \vspace{0.1cm} \\'
                                      % (cfg.vollbl, cfg.report_date.date().strftime("%Y-%-m-%-d")) +
                                      spacecmd + swe_byelev.to_latex(na_rep='-', column_format=colstr) +
                                      r'} \\ \footnotesize{\textbf{Table %s:} SWE volume.}%s' % (str(mtables), clrpage))
        mtables += 1

    else:
        variables['SWEVOL_BYELEV'] = ''

    if 'swe_change' in cfg.tables:
        ptables += 1

        if ptables == 2:
            clrpage = r'\clearpage'
        else:
            clrpage = ''
        dbval = collect(cnx, plotorder, basins, start_date, start_date, 'swe_z', run_name, dfindt, 'end')
        dbval = dbval.rename(columns=cfg.labels)
        start_swe = dbval.round(decimals=dpts)
        dbval = collect(cnx, plotorder, basins, start_date, end_date, 'swe_z', run_name, dfindt, 'end')
        dbval = dbval.rename(columns=cfg.labels)
        end_swe = dbval.round(decimals=dpts)
        dswe_byelev = end_swe - start_swe
        dswe_byelev.rename(index={'total': 'mean'}, inplace=True)
        dswe_byelev.index.name = 'Elevation'
        variables['DSWE_BYELEV'] = (r'  \normalsize \textbf{Change in SWE [%s], %s to %s}\\ \vspace{0.1cm} \\'
                                    % (cfg.depthlbl, cfg.report_start.date().strftime("%Y-%-m-%-d"),
                                       cfg.report_date.date().strftime("%Y-%-m-%-d")) + spacecmd +
                                    dswe_byelev.to_latex(na_rep='-', column_format=colstr) +
                                    r'} \\ \footnotesize{\textbf{Table %s:} Change in SWE.} %s' % (
                                    str(mtables), clrpage))
        mtables += 1

    else:
        variables['DSWE_BYELEV'] = ''

    if 'swe_percent' in cfg.tables:
        ptables += 1
        if (ptables % 2) == 0 and ptables != 0:
            clrpage = r'\clearpage'
        else:
            clrpage = ''

        dbval = collect(cnx, plotorder, basins, start_date, end_date, 'swe_vol', run_name, dfind, 'end')
        dbval = dbval.rename(columns=cfg.labels)
        swe_byelev = dbval.round(decimals=dpts)
        value = swe_byelev.iloc[:-1].sum()
        sweper_byelev = (swe_byelev / value * 100).round(decimals=dpts)
        sweper_byelev.index.name = 'Elevation'

        variables['SWEPER_BYELEV'] = (
                    r'  \normalsize \textbf{SWE volume, percent of basin total, %s}\\ \vspace{0.1cm} \\'
                    % (cfg.report_date.date().strftime("%Y-%-m-%-d")) +
                    spacecmd + sweper_byelev.round(1).to_latex(na_rep='-', column_format=colstr) +
                    r'}  \\ \footnotesize{\textbf{Table %s:} Percent of total SWE volume.}%s' % (str(mtables), clrpage))
        variables['SWEPER_BYELEV'] = variables['SWEPER_BYELEV'].replace('inf', '-')

        mtables += 1

    else:
        variables['SWEPER_BYELEV'] = ''

    if 'swi_vol' in cfg.tables:
        dbval = collect(cnx, plotorder, basins, start_date, end_date, 'swi_vol', run_name, dfindt, 'sum')
        dbval = dbval.rename(columns=cfg.labels)
        swi_byelev = dbval.round(decimals=dpts)
        variables['ACCUM_BYELEV'] = (r' \normalsize \textbf{SWI [%s] by elevation, %s to %s}\\ \vspace{0.1cm} \\'
                                     % (cfg.vollbl, cfg.report_start.date().strftime("%Y-%-m-%-d"),
                                        cfg.report_date.date().strftime("%Y-%-m-%-d")) + spacecmd +
                                     swi_byelev.to_latex(na_rep='-', column_format=colstr) +
                                     r'} \\ \footnotesize{\textbf{Table %s:} SWI volume. }' % (str(mtables)))
        mtables += 1

    else:
        variables['ACCUM_BYELEV'] = ''

    variables['TOT_LBL'] = cfg.plotorder[0]

    # for n in range(1,len(cfg.plotorder)):
    for n in range(1, len(cfg.labels)):
        s = 'SUB' + str(n) + '_LBL'
        variables[s] = cfg.labels[cfg.plotorder[n]]

    # Convert floats to strings
    for name in variables:
        if isinstance(variables[name], float):
            if cfg.dplcs == 0:
                tmp = str(int(variables[name]))
            else:
                tmp = str(round(variables[name], cfg.dplcs))
            variables[name] = tmp

    # Summary sections and fig template have variable strings
    # (e.g. CHANGES_FIG) that need to be replaced
    section_dict = {'SUMMARY': cfg.summary_file,
                    'CHANGES_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'changes_fig_tpl.txt'),
                    'SWI_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'swi_fig_tpl.txt'),
                    'TOTALS_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'totals_fig_tpl.txt'),
                    'MULTITOTSWE_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'multitotswe_fig_tpl.txt'),
                    'VALID_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'valid_fig_tpl.txt'),
                    'FLTCHANGES_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'flt_fig_tpl.txt'),
                    'PDEP_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'pdep_fig_tpl.txt'),
                    'COLD_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'cold_fig_tpl.txt'),
                    'SWE_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'swe_fig_tpl.txt'),
                    'SUBBASINS_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'subbasins_fig_tpl.txt'),
                    'SWEFORECAST_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'forecastswe_fig_tpl.txt'),
                    'SWIFORECAST_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'forecastswi_fig_tpl.txt'),
                    'CHANGESFORECAST_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'forecastchanges_fig_tpl.txt'),
                    'TOTALSFORECAST_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'forecasttotals_fig_tpl.txt'),
                    'PDEPFORECAST_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'forecastpdep_fig_tpl.txt'),
                    'DIAGNOSTICS_FIG_TPL': os.path.join(cfg.figs_tpl_path, 'diagnostics_fig_tpl.txt')
                    }

    # Define and load summary tables depending on number of subbasins
    section_dict['SWE_SUMMARY_TPL'] = os.path.join(cfg.figs_tpl_path,
                                                   'swe_summary_{}sub.txt'.format(str(len(cfg.plotorder))))

    # Remove if no flight
    if not cfg.flt_flag or not cfg.flight_figs:
        del section_dict['FLTCHANGES_FIG_TPL']

    if not cfg.report_diagnostics:
        del section_dict['DIAGNOSTICS_FIG_TPL']
        variables['DIAGNOSTICS_FIG'] = ''

    # Remove if no forecast
    if not cfg.forecast_flag:
        del section_dict['SWEFORECAST_FIG_TPL']
        del section_dict['SWIFORECAST_FIG_TPL']
        del section_dict['CHANGESFORECAST_FIG_TPL']
        del section_dict['TOTALSFORECAST_FIG_TPL']
        del section_dict['PDEPFORECAST_FIG_TPL']

        variables['SWIFORECAST_FIG'] = ''
        variables['SWIFORECAST_FIG_TPL'] = ''
        variables['PDEPFORECAST_FIG'] = ''
        variables['PDEPFORECAST_FIG_TPL'] = ''
        variables['SWEFORECAST_FIG'] = ''
        variables['SWEFORECAST_FIG_TPL'] = ''
        variables['CHANGESFORECAST_FIG'] = ''
        variables['CHANGESFORECAST_FIG_TPL'] = ''
        variables['TOTALSFORECAST_FIG'] = ''
        variables['TOTALSFORECAST_FIG_TPL'] = ''
        variables['FORE_TITLE'] = ''

    else:
        variables['FORE_TITLE'] = 'with WRF forecast, {} to {}'.format(cfg.for_start_date.date().strftime("%b %-d"),
                                                                       cfg.for_end_date.date().strftime("%b %-d"))

    for rep in section_dict.keys():

        # for variable numbers of fligth figures
        if rep == 'FLTCHANGES_FIG_TPL':
            fid = open(section_dict[rep], 'r')
            var = fid.read()
            fid.close()

            for name in sorted(variables):
                if name == 'DFLT_FIG':
                    for i, fltname in enumerate(cfg.flight_diff_fig_names):
                        flt_date = cfg.flight_outputs['dates'][i].date().strftime("%Y%m%d")
                        flt_num = cfg.flight_delta_vol_df[flt_date].round(1).to_latex(na_rep='-', column_format=colstr)
                        table = (r'{ \vspace{0.5cm} \textbf{Change in SWE [%s] by elevation, ' % (cfg.vollbl) +
                                 r'from %s update} \\ \vspace{0.1cm} \\' % (flt_date) +
                                 r' %s %s }' % (spacecmd, flt_num))

                        if i == 0:
                            tmp = var.replace(name, fltname)
                        else:
                            tmp += var.replace(name, fltname)

                        tmp += table

                    var = tmp.replace(name, variables[name])

                else:
                    var = var.replace(name, variables[name])

            variables[rep] = var

        else:
            fid = open(section_dict[rep], 'r')
            var = fid.read()
            fid.close()

            for name in sorted(variables):
                var = var.replace(name, variables[name])
            variables[rep] = var

    if not cfg.subs_fig:
        variables['SUBBASINS_FIG_TPL'] = ''

    if not cfg.rep_compare_runs_flag:
        variables['MULTITOTSWE_FIG_TPL'] = ''

    if not cfg.rep_swi_flag:
        variables['SWI_FIG_TPL'] = ''

    if not cfg.rep_image_change_flag:
        variables['CHANGES_FIG_TPL'] = ''

    if not cfg.rep_cold_content_flag:
        variables['COLD_FIG_TPL'] = ''

    if not cfg.rep_swe_volume_flag:
        variables['SWE_FIG_TPL'] = ''

    if not cfg.rep_basin_total_flag:
        variables['TOTALS_FIG_TPL'] = ''

    if not cfg.rep_stn_validate_flag:
        variables['VALID_FIG_TPL'] = ''

    if not cfg.rep_compare_runs_flag:
        variables['MULTITOTSWE_FIG_TPL'] = ''

    if not cfg.rep_precip_depth_flag:
        variables['PDEP_FIG_TPL'] = ''

    # Make the report
    env = make_env(loader=FileSystemLoader(cfg.templ_path))
    tpl = env.get_template(cfg.tex_file)

    # print VAR-replaced latex file for debugging if desired
    if cfg.print_latex:
        print(tpl.render(variables))

    pdf = build_pdf(tpl.render(variables))

    # Save in reports and with figs
    rpath = os.path.join(cfg.figs_path, '' + cfg.report_name)
    pdf.save_to(rpath)
    cfg._logger.info(' Saved {}'.format(rpath))

    if cfg.rep_path is not None:
        rpath = os.path.join(cfg.rep_path, '' + cfg.report_name)
        pdf.save_to(rpath)
        cfg._logger.info(' Saved {}'.format(rpath))




import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns
from datetime import datetime
import pandas as pd
from matplotlib.dates import DateFormatter
import dateutil.parser
import copy
import snowav.framework.figures


def basin_total(args, logger = None):
    '''
    Basin total SWE and SWI.

    Args
    ----------
    args : dict
        dictionary with required inputs, see swi() figure for more information.

    logger : dict
        snowav logger

    '''

    plotorder = args['plotorder']
    labels = args['labels']
    barcolors = args['barcolors']
    barcolors.insert(0,'black')
    swi_summary = args['swi_summary']
    swe_summary = args['swe_summary']
    swi_end_val = copy.deepcopy(swi_summary)
    swe_title = 'Basin SWE'
    swi_title = 'Basin SWI'

    sns.set_style('darkgrid')
    sns.set_context("notebook")

    plt.close(8)
    fig,(ax,ax1) = plt.subplots(num=8, figsize=args['figsize'],
                                dpi=args['dpi'], nrows = 1, ncols = 2)

    for iters,name in enumerate(plotorder):
        swe_summary[name].plot(ax=ax, color = barcolors[iters],
                               label = labels[name])
        swi_summary[name].plot(ax=ax1,color = barcolors[iters],
                               label='_nolegend_')

    if args['flt_flag']:
        for i,d in enumerate(args['flight_dates']):
            if i == 0:
                lb = 'flight update'.format(args['wy'])

            else:
                lb = '__nolabel__'

            ax.axvline(x=d, linestyle=':', linewidth=0.75, color='k', label=lb)

    x_end_date = args['end_date']

    # forecast
    # if args['forecast_flag']:
        # print('basin_total forecast in progress')
        # ax.axvline(x=snow.for_start_date,
        #            linestyle = ':',
        #            linewidth = 0.75,
        #            color = 'r')
        # ax1.axvline(x=snow.for_start_date,
        #            linestyle = ':',
        #            linewidth = 0.75,
        #            color = 'r')

    ax1.yaxis.set_label_position("right")
    ax1.set_xlim((datetime(args['wy']-1, 10, 1), x_end_date))
    ax.set_xlim((datetime(args['wy']-1, 10, 1), x_end_date))
    ax1.tick_params(axis='y')
    ax1.yaxis.tick_right()
    ax.legend(loc='upper left')

    # Put on the same yaxis
    swey = ax.get_ylim()
    swiy = ax1.get_ylim()

    if swey[1] < swiy[1]:
        ax1.set_ylim((-0.1,swiy[1]))
        ax.set_ylim((-0.1,swiy[1]))

    if swey[1] >= swiy[1]:
        ax1.set_ylim((-0.1,swey[1]))
        ax.set_ylim((-0.1,swey[1]))

    for tick,tick1 in zip(ax.get_xticklabels(),ax1.get_xticklabels()):
        tick.set_rotation(30)
        tick1.set_rotation(30)

    ax1.set_ylabel(r'[{}]'.format(args['vollbl']))
    ax1.set_xlabel('')
    ax.set_xlabel('')
    ax.axes.set_title(swe_title)
    ax1.axes.set_title(swi_title)
    ax.set_ylabel(r'[{}]'.format(args['vollbl']))

    del barcolors[0]

    fig_name_short = 'basin_total_'
    fig_name = '{}{}{}.png'.format(args['figs_path'],fig_name_short,args['directory'])
    if logger is not None:
        logger.info(' saving {}'.format(fig_name))
    snowav.framework.figures.save_fig(fig, fig_name)

    return fig_name_short




import pandas as pd
import os
from datetime import datetime
import copy
import seaborn as sns
from matplotlib import pyplot as plt
import snowav.framework.figures

def inflow(args, logger):
    '''
    Plot inflow and SWI.

    Currently this is only configured for the Tuolumne and San Joaquin, and
    requires some manipulation of csv files.

    Used openoffice terminal command to create original csv file from xlsb
    file from USBR:

        $ soffice --headless --convert-to csv SJ.xlsb

    pyxlsb also a possibility:

        from pyxlsb import open_workbook as open_xlsb
        df = []
        wb = open_xlsb('<.xlsb>')
        for n in range(1,10):
            sheet = wb.get_sheet(n)
            for row in sheet.rows():
                df.append([item.v for item in row])

    manually assigning inflow from sheets, see sample at
    /home/markrobertson/wkspace/projects/inflow/sanjoaquin.csv

    Args
    ----------
    args : dict
        dictionary with required inputs, see swi() figure for more information.

    logger : list
        snowav logger

    '''

    inflow_summary = args['inflow_summary']
    inflow_headings = args['inflow_headings']
    basin_headings = args['basin_headings']
    inflow_summary = inflow_summary.cumsum()

    plotorder = args['plotorder']
    labels = args['labels']
    barcolors = args['barcolors']
    barcolors.insert(0,'black')
    swi_summary = args['swi_summary']
    swi_end_val = copy.deepcopy(swi_summary)
    swi_title = 'SWI and Reservoir Inflow'

    sns.set_style('darkgrid')
    sns.set_context("notebook")

    plt.close(0)
    f = plt.figure(num=0, figsize=(8,5), dpi=args['dpi'])
    a = plt.gca()

    for iters, (name, basin) in enumerate(zip(basin_headings, inflow_headings)):
        if iters == 0:
            lblswi = 'SWI'
            lblin = 'inflow'
        else:
            lblswi = '__nolabel__'
            lblin = '__nolabel__'

        swi_summary[name].plot(ax = a, color = barcolors[iters],
                               linestyle=':', linewidth = 0.8, label = lblswi)
        inflow_summary[name].plot(ax = a, color = barcolors[iters],
                                  label = lblin)

    x_end_date = args['end_date']

    a.set_xlim((datetime(args['wy']-1, 10, 1), x_end_date))
    a.legend(loc = 'upper left')

    for tick in a.get_xticklabels():
        tick.set_rotation(30)

    a.set_xlabel('')
    a.axes.set_title(swi_title)
    a.set_ylabel(r'[{}]'.format(args['vollbl']))

    del barcolors[0]

    fig_name_short = 'inflow_'
    fig_name = '{}{}{}.png'.format(args['figs_path'],fig_name_short,args['directory'])
    if logger is not None:
        logger.info(' saving {}'.format(fig_name))

    snowav.framework.figures.save_fig(f, fig_name)

    return fig_name_short

def excel_to_csv(args, logger):
    '''
    Read in summary inflow data from operators, pull full natural flow
    calculations, and write out summary csv file.

    Args
    ------
    args : dict
    logger : object

    '''

    path = args['path']
    csv_file = args['csv_file']
    basin_headings = args['basin_headings']
    inflow_headings = args['inflow_headings']
    file_base = args['file_base']
    sheet_name = args['sheet_name']
    skiprows = args['skiprows']
    date_idx = args['date_idx']
    wy = args['wy']
    overwrite = args['overwrite']
    convert = args['convert']

    if csv_file is None or not os.path.isfile(csv_file):
        if logger is not None:
            logger.info(' {} not a file, creating it...'.format(csv_file))

        date_range = pd.date_range(datetime(wy-1,10,1),datetime(wy,9,30),freq='D')
        csv = pd.DataFrame(index = date_range, columns = basin_headings)

    else:
        csv = pd.read_csv(csv_file, parse_dates=[0], index_col = 0)

    # Tuolumne
    if inflow_headings[0] == 'HETCHY':
        date_heading = '        HETCH HETCHY WATER AND POWER SYSTEM'

        # if no path is supplied we just read in existing csv
        if path is not None:
            files = os.listdir(path)

            for file in files:
                if file_base in file:
                    # get date from loading in full file
                    f = pd.read_excel(os.path.join(path,file),
                                      sheet_name = sheet_name,
                                      skiprows = 0)
                    inflow_date = pd.to_datetime(f[date_heading].values[date_idx])
                    data = pd.read_excel(os.path.join(path,file),
                                         sheet_name=sheet_name,
                                         skiprows=skiprows)

                    if not inflow_date.date() in csv.index or overwrite:
                        for basin,heading in zip(basin_headings,inflow_headings):
                            if logger is not None:
                                logger.debug(' Assigning basin inflow in {}: {}, {} to '
                                              '{}'.format(file, inflow_date, heading, basin))
                            csv.loc[inflow_date,basin] = data[heading].values[11]*convert

        csv.sort_index(inplace=True)
        csv.to_csv(csv_file)


    # if inflow_headings[0] == 'Millerton':
    #     inflow_headings[0] = 'Millerton Inflow'
    #
    #     date_heading = 'date'
    #     files = os.listdir(path)
    #     f = pd.read_csv(os.path.join(path,files[0]), index_col = 'date')
    #     f = f['Millerton Inflow']*convert
    #
    #     f.to_csv(csv_file)



from datetime import datetime
from itertools import count, filterfalse
import mysql.connector
import numpy as np
import os
import pandas as pd
import sqlalchemy as sa
from sqlalchemy import create_engine, and_
from sqlalchemy.orm import sessionmaker
from sys import exit
import urllib.parse
import warnings

from snowav.database import tables as ta
import snowav
from snowav.database.tables import Base, RunMetadata, Watershed, Basin, \
    Results, VariableUnits

# Fix these two by pulling smrf and awsm versions from netcdf
try:
    import smrf

    smrf_version = smrf.__version__
except:
    smrf_version = 'unknown'

try:
    warnings.filterwarnings("ignore")
    import awsm

    warnings.filterwarnings("default")
    awsm_version = awsm.__version__
except:
    awsm_version = 'unknown'


class Database(object):
    """ Database class for snowav.

    Args
    ------
    db_type {string}: database type
    db_name {string}: mysql database name, if using db_type='sql'
    user {string}: database user, if using db_type='sql'
    password {string}: database password, if using db_type='sql'
    host {string}: database host, if using db_type='sql'
    port {string}: database port, if using db_type='sql'
    database {string}: sqlite database file path, if using db_type='sqlite'
    """

    def __init__(self, db_type='sql', db_name='snowav', user=None,
                 password=None, host=None, port=None, database=None):

        db_type_options = ['sql', 'sqlite']
        if db_type not in db_type_options:
            raise ValueError('Database input "db_type" must be one of '
                             '{}'.format(db_type_options))

        if db_type == 'sql':
            input_list = [user, password, host, port]
            if sum([x is not None for x in input_list]) != len(input_list):
                raise Exception('Database input db_type "sql" requires '
                                '"user", "password", "host", and "port"')

        if db_type == 'sqlite' and database is None:
            raise Exception('must supply "database" with db_type=sqlite')

        self.db_type = db_type
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.db_name = db_name
        self.database = database

    def assign_vars(self, kwargs):
        """ Assign Database attributes.

        Args
        ------
        kwargs {dict}: {item, value}
        """

        if not isinstance(kwargs, dict):
            raise TypeError("kwargs must be dict")

        for item, v in kwargs.items():
            setattr(self, item, v)

    def make_connection(self, logger=None):
        """ Make and check database connection.

        Assigns self.engine

        Args
        ------
        logger {class}: logger
        """

        if self.db_type == 'sql':
            connector = '{}{}:{}@{}:{}/{}'.format('mysql+mysqlconnector://',
                                                  self.user,
                                                  self.password,
                                                  self.host,
                                                  self.port,
                                                  self.db_name)
            if logger is not None:
                logger.info(" Using database connection "
                            "{}@{}:{}".format(self.user, self.host, self.port))

        if self.db_type == 'sqlite':
            connector = 'sqlite:///' + self.database
            if logger is not None:
                logger.info(" Using database connector: {}".format(connector))

        try:
            engine = create_engine(connector)
        except Exception as e:
            print(e)

        self.assign_vars({'engine': engine})

    def check_tables(self, logger=None):
        """ Make database tables.

        Args
        ------
        logger {class}: logger
        """

        if not hasattr(self, 'engine'):
            if logger is not None:
                logger.error(" Database connector required...")
            else:
                print('Database connector required')

        Base.metadata.create_all(self.engine)

    def insert(self, dbtable, kwargs, logger=None):
        """ Put data on database.

        Args
        ------
        dbtable {string}: string format of database table name (i.e., 'Pixels')
        kwargs {dict}: data to put on the database {field: value}
        logger {class}: logger
        """

        if type(kwargs) != dict:
            raise TypeError("kwargs must be a dict")

        # get database table
        tbl = sa.Table(dbtable, sa.MetaData(), autoload_with=self.engine)

        # table columns
        fields = list(tbl.columns.keys())

        for key in kwargs:
            if key not in fields:
                if logger is not None:
                    logger.warning(" kwargs field '{}' not in table columns "
                                   "{}".format(key, fields))
                else:
                    print("query() kwargs field '{}' not in table columns "
                          "{}".format(key, fields))

        with self.engine.connect() as dbcon:
            dbcon.execute(tbl.insert(), kwargs)

    def operators(self, operator):
        """ Translate sqlalchemy query operators for readability.

            eq for ==
            lt for <
            ge for >=
            in for in_
            like for like

        """

        if not isinstance(operator, str):
            raise TypeError("operators must be a str")

        omap = {'==': 'eq',
                '<': 'lt',
                '>': 'gt',
                '>=': 'ge',
                '<=': 'le',
                'in_': 'in',
                'like': 'like'}

        if operator not in list(omap.keys()):
            raise ValueError("operator must be in "
                             "'{}'".format(list(omap.keys())))

        return omap[operator]

    def query(self, params, columns=None, index=None, logger=None):
        """ Query database values.

        Columns and index are used to filter results after query. Columns
        is applied before index, so if you want index='index', 'index' must
        also appear in columns=['index'].

        Args
        ------
        params {dict}: query parameters, in the format
                        {table: (column, operator, value)}
        columns {list}: list of database table columns to filter results
        index {list}: name of database table column to set as returned
            DataFrame index
        logger {class}: logger

        Returns
        ------
        results {DataFrame}: query results
        """

        # tparams = {Pixels: {(Pixels.model_row, '==', int(pt[0]))}}

        unique_tables = set(params.keys())
        tables = list(params.keys())
        ntables = len(unique_tables)

        if ntables < 1 or ntables > 2:
            raise ValueError("params must have <= 2 unique tables")

        dbsession = sessionmaker(bind=self.engine)
        session = dbsession()

        from snowav.database.tables import Pixels, PixelsData

        n = 0
        for t in params.keys():
            print(t)
            if t == 'Pixels':
                table = Pixels
            if t == 'PixelsData':
                table = PixelsData

            for f in params[t]:
                print(f)
                raw = f

                try:
                    key, op, value = raw
                except ValueError:
                    raise Exception('Invalid filter: %s' % raw)

                column = getattr(table, key, None)

                if not column:
                    raise Exception('Invalid filter column: %s' % key)
                if op == 'in':
                    if isinstance(value, list):
                        filt = column.in_(value)
                    else:
                        filt = column.in_(value.split(','))
                else:
                    try:
                        attr = list(filter(
                            lambda e: hasattr(column, e % op),
                            ['%s', '%s_', '__%s__']
                        ))[0] % op
                    except IndexError:
                        raise Exception('Invalid filter operator: %s' % op)
                    if value == 'null':
                        value = None

                    if n == 0:
                        filt = getattr(column, attr)(value)
                        print("FIRST\n", filt)
                    else:
                        t = getattr(column, attr)(value)
                        filt = and_(filt, t)
                        print('AND\n ', filt)

                n += 1

        if ntables == 1:
            qry = session.query(table).filter(filt)
        else:
            qry = session.query(tables[0]).join(tables[1])

        # qry = qry.filter(filt)

        print('QRY\n', qry)
        results = pd.read_sql(qry.statement, qry.session.connection())
        session.close()

        print(results)
        print(x)

        """
        if index is None:
            index = []
        if columns is None:
            columns = []
        if not isinstance(params, dict):
            raise TypeError("params must be a dict")

        unique_tables = set(params.keys())
        ntables = len(unique_tables)

        if ntables < 1 or ntables > 2:
            raise ValueError("params can have no more than 2 unique tables")

        if not isinstance(columns, list):
            raise TypeError("columns must be a list")

        if not isinstance(index, list):
            raise TypeError("index must be a list")

        dbsession = sessionmaker(bind=self.engine)
        session = dbsession()
        tables = list(params.keys())
        ntables = len(tables)

        if ntables == 1:
            qry = session.query(tables[0]).filter_by(**params[tables[0]])
        else:
            qry = session.query(tables[0], tables[1])

        results = pd.read_sql(qry.statement, qry.session.connection())
        session.close()

        # apply columns
        applied_columns = []
        for c in columns:
            if c in results.columns:
                applied_columns.append(c)

            else:
                if logger is not None:
                    logger.warning(" columns='{}' not in {}, will not "
                                   "apply".format(c, list(results.columns)))
                else:
                    print("columns='{}' not in {}, will not "
                          "apply".format(c, list(results.columns)))

        if len(applied_columns) > 0:
            results = results[applied_columns]

        # apply index
        if len(index) > 0:
            if index[0] not in list(results.columns):
                if logger is not None:
                    logger.warning(" index='{}' not in columns={}, will not "
                                   "apply".format(index[0],
                                                  list(results.columns)))
                else:
                    print("index='{}' not in columns={}, will not "
                          "apply".format(index[0], list(results.columns)))
            else:
                results.set_index(index, inplace=True)
                results.sort_index(inplace=True)

        return results
        """

    def delete(self, dbtable, kwargs, logger=None):
        """ Delete database records.

        Args
        ------
        dbtable {string}: string format of database table name (i.e., 'Pixels')
        kwargs {dict}: items for deletion {field: value}
        logger {class}: logger
        """

        if not isinstance(kwargs, dict):
            raise TypeError("kwargs must be dict")

        dbsession = sessionmaker(bind=self.engine)
        session = dbsession()

        tbl = sa.Table(dbtable, sa.MetaData(), autoload_with=self.engine)
        results = session.query(tbl).filter_by(**kwargs)
        results.delete(synchronize_session=False)

        session.commit()
        session.close()

        if logger is not None:
            logger.debug(" Deleted database records")


def make_session(connector):
    '''
    Make snowav database session.

    Args
    -------
    connector : str
        database connector, either path to sqlite database or mysqlconnector
        string

    Returns
    -------
    session : object
        snowav database session
    '''

    try:
        engine = create_engine(connector)
    except:
        raise Exception('Failed to make database connection with '
                        '{}'.format(connector))
    try:
        Base.metadata.create_all(engine)
    except:
        raise Exception('Failed establishing Base for sqlalchemy')

    DBSession = sessionmaker(bind=engine)
    session = DBSession()

    return session


def insert(connector, table, values):
    '''
    Inserts standard run results to the database.

    Args
    --------
    session : object
        database session
    table : str
        database table (RunMetadata, Results, or VariableUnits)
    values : dict
        dictionary of values

    '''

    dbtable = getattr(snowav.database.tables, table)
    my_dbtable = dbtable()

    for k, v in values.items():
        setattr(my_dbtable, k, v)

    session = make_session(connector)
    session.add(my_dbtable)
    session.commit()
    session.close()


def collect(connector, plotorder, basins, start_date, end_date, value,
            run_name, edges, method):
    '''

    Collect snowav database results into our standard dataframe with elevation
    rows and subbasin columns.

    Args
    ---------
    session : object
        sqlalchemy snowav database session
    plotorder : str
        list of basins
    start_date : datetime
    end_date : datetime
    value : str
        database value identifier
    run_name : str
        database run name identifier
    edges : list
        elevation band, can include 'total'
    method : str
        options [end sum difference daily]

    Returns
    ----------
    df : DataFrame
        standard subbasin columns and elevation labels

               Extended Tuolumne  Tuolumne  Cherry Creek  Eleanor
        3000               0.000     0.000         0.000    0.000
        ...                  ...       ...           ...      ...

    '''

    value_options = ['swe_z', 'swe_vol', 'density', 'precip', 'precip_z', 'precip_vol',
                     'rain_z', 'rain_vol', 'swe_avail', 'swe_unavail', 'coldcont',
                     'swi_z', 'swi_vol', 'depth', 'snow_line', 'mean_air_temp',
                     'evap_z', 'L_v_E', 'melt', 'lwc', 'temp_surface',
                     'temp_lower', 'temp_bulk', 'depth_lower_layer', 'h20_sat',
                     'R_n', 'H', 'G', 'M', 'delta_Q']

    method_options = ['sum', 'difference', 'daily', 'end']

    if value not in value_options:
        raise Exception('value={} call to database.collect() '.format(value) +
                        'must be one of {}'.format(value_options))

    if method not in method_options:
        raise Exception('method={} call to database collect() '.format(method) +
                        'must be one of {}'.format(method_options))

    if type(plotorder) != list:
        plotorder = [plotorder]

    if edges == 'total':
        edges = ['total']

    if method == 'daily':
        df = pd.DataFrame(index=['date_time'], columns=plotorder)

    else:
        df = pd.DataFrame(np.nan, index=edges, columns=plotorder)

    for bid in plotorder:
        results = query(connector, start_date, end_date, run_name, basins, bid, value)

        if results.empty:
            raise Exception('No results on database for '
                            'run_name: {}, '
                            'start_date: {}, '
                            'end_date: {}, '
                            'value: {} '.format(run_name, start_date,
                                                end_date, value))

        if method == 'daily':
            e = results[(results['elevation'] == 'total')
                        & (results['date_time'] >= start_date)
                        & (results['date_time'] <= end_date)]
            if e.empty:
                raise Exception('No results on database for '
                                'run_name: {} ',
                                'start_date: {} ,'
                                'end_date: {}, '
                                'value: {} ,'
                                'elevation: {}'.format(run_name, start_date,
                                                       end_date, value, elev))
            else:
                e = e.set_index('date_time')
                e.sort_index(inplace=True)

                if bid == plotorder[0]:
                    df = e[['value']].copy()
                    df = df.rename(columns={'value': bid})

                else:
                    df[bid] = e['value']

            df.sort_index(inplace=True)

        if method == 'end':
            for elev in edges:
                e = results[(results['elevation'] == str(elev))
                            & (results['date_time'] == end_date)]
                if e.empty:
                    raise Exception('Empty results for database query in '
                                    'database.collect(), for run_name={}, '
                                    'elev={}, {} to {}'.format(run_name, elev,
                                                               start_date, end_date))
                else:
                    if e['value'].values[0] is None:
                        df.loc[elev, bid] = np.nan
                    else:

                        # The database can occasionally get multiple values
                        # for the same record if it doesn't exit cleanly
                        if len(e['value'].values) == 1:
                            df.loc[elev, bid] = e['value'].values
                        else:
                            raise Exception('Multiple database entries for a '
                                            'single field, consider running with '
                                            '[database] overwrite: True')

        if method == 'difference':
            for elev in edges:
                s = results[(results['elevation'] == str(elev))
                            & (results['date_time'] == start_date)]
                e = results[(results['elevation'] == str(elev))
                            & (results['date_time'] == end_date)]
                if e.empty or s.empty:
                    raise Exception('Empty results for database query in '
                                    'database.collect(), for run_name={}, '
                                    'elev={}, {} to {}'.format(run_name, elev,
                                                               start_date, end_date))
                else:
                    df.loc[elev, bid] = np.nansum(e['value'].values - s['value'].values)

        if method == 'sum':
            for elev in edges:
                e = results[(results['elevation'] == str(elev))
                            & (results['date_time'] >= start_date)
                            & (results['date_time'] <= end_date)]
                if e.empty:
                    raise Exception('Empty results for database query in '
                                    'database.collect(), for run_name={}, '
                                    'elev={}, {} to {}'.format(run_name, elev,
                                                               start_date, end_date))
                else:
                    df.loc[elev, bid] = e['value'].sum(skipna=False)

    df = df[plotorder]

    return df


def query(connector, start_date, end_date, run_name, basins, bid=None,
          value=None, rid=None):
    '''
    Retrieve results from snowav database.

    Args
    --------
    start_date : datetime
    end_date : datetime
    run_name : str
        identifier for run, specified in config file
    bid : str
        basin id in string format ('Boise River Basin')
    value : str
        value to query (i.e. 'swi_z')
    rid : run_id

    Returns
    ---------
    df : DataFrame
        dataframe of query results

    '''

    basin_id = int(basins[bid]['basin_id'])
    session = make_session(connector)

    if (value != None) and (bid != None):
        qry = session.query(Results).join(RunMetadata).filter(and_(
            (Results.date_time >= start_date),
            (Results.date_time <= end_date),
            (RunMetadata.run_name == run_name),
            (Results.variable == value),
            (Results.basin_id == basin_id)))

    elif (value == None) and (bid != None) and rid != None:
        qry = session.query(Results).join(RunMetadata).filter(and_(
            (Results.date_time >= start_date),
            (Results.date_time <= end_date),
            (Results.run_id == int(rid)),
            (RunMetadata.run_name == run_name),
            (Results.basin_id.in_(bids))))

    else:
        qry = session.query(Results).join(RunMetadata).filter(and_(
            (Results.date_time >= start_date),
            (Results.date_time <= end_date),
            (RunMetadata.run_name == run_name)))

    df = pd.read_sql(qry.statement, qry.session.connection())

    session.close()

    return df


def delete(connector, basins, start_date, end_date, bid, run_name):
    '''
    Delete results from the database. This deletes values with basin_id == bid
    and run_name = run_name within the date range. If there are no more
    remaining records for that run_id (the entire run was deleted), then the
    RunMetadata and VariableUnits data are also removed.

    Note: deletion order matters because of table relationships.

    Args
    --------
    start_date : datetime
    end_date : datetime
    bid : str
        basin name
    run_name : str
        identifier for run, specified in config file

    Returns
    --------
    logger : list

    '''

    logger = []
    basin_id = int(basins[bid]['basin_id'])
    session = make_session(connector)

    logger.append(' Deleting existing records for {}, {}, {} '.format(
        bid, run_name, start_date.date()))

    # Get the run_id
    qry = session.query(Results).join(RunMetadata).filter(and_(
        (Results.date_time >= start_date),
        (Results.date_time <= end_date),
        (RunMetadata.run_name == run_name),
        (Results.basin_id == basin_id)))

    df = pd.read_sql(qry.statement, qry.session.connection())
    unique_runs = df.run_id.unique()

    if not df.empty:
        for r in unique_runs:
            session.query(Results).filter(and_((Results.date_time >= start_date),
                                               (Results.date_time <= end_date),
                                               (Results.run_id == int(r)),
                                               (Results.basin_id == basin_id))).delete()

            session.commit()

        # Query again, if no results from those run_id exist still, also remove
        # the metadata and variableUnits
        qrym = session.query(Results).filter(Results.run_id == int(r))
        dfm = pd.read_sql(qrym.statement, qrym.session.connection())

        if dfm.empty:
            logger.append(' Deleting RunMetadata run_name={}, run_id={}, from {} '
                          'to {}'.format(run_name, str(r), start_date.date(),
                                         end_date.date()))

            session.query(VariableUnits).filter(VariableUnits.run_id == int(r)).delete()
            session.query(RunMetadata).filter(RunMetadata.run_id == int(r)).delete()
            session.commit()

    session.close()

    return logger


def create_tables(database, plotorder):
    '''
    Creates database Watersheds and Basins tables. This is currently only
    being called on database creation.

    Args
    -------
    database : str
        database identifier
    plotorder : list

    Returns
    --------
    logger : list


    '''

    logger = []

    engine = create_engine(database)
    Base.metadata.create_all(engine)
    DBSession = sessionmaker(bind=engine)
    session = DBSession()

    logger.append(' Adding watershed {} to database'.format(plotorder[0]))

    # Initialize watershed
    wval = Watershed(watershed_id=1, watershed_name=plotorder[0])
    session.add(wval)
    session.commit()

    # Initialize basins within the watershed
    for i, name in enumerate(plotorder):
        logger.append(' Adding basin {} to database'.format(name))

        bval = Basin(watershed_id=1,
                     basin_id=i + 1,
                     basin_name=name)

        session.add(bval)

    session.commit()
    session.close()

    logger.append(' Completed initializing basins and watersheds')

    return logger


def run_metadata(cfg):
    '''
    Create database RunMetadata for each snowav run.

    Args
    ------
        Variables: class, from snowav/database/models.py
        run_name : str

    '''

    watershed_id = int(cfg.basins[cfg.plotorder[0]]['watershed_id'])

    session = make_session(cfg.connector)
    qry = session.query(RunMetadata)
    df = pd.read_sql(qry.statement, qry.session.connection())
    session.close()

    # Increase each runid by 1
    if not df.run_id.values.size:
        cfg.run_id = 1
    else:
        r = df['run_id'].values
        cfg.run_id = next(filterfalse(set(r).__contains__, count(1)))

    # in most cases it's not practical to save all run_dirs
    trdir = ','.join(cfg.run_dirs)
    if len(trdir) > 1200:
        trdir = cfg.run_dirs[0]

    values = {'run_id': int(cfg.run_id),
              'run_name': cfg.run_name,
              'watershed_id': int(watershed_id),
              'pixel': int(cfg.pixel),
              'description': '',
              'smrf_version': 'smrf' + smrf_version,
              'awsm_version': 'aswm' + awsm_version,
              'snowav_version': 'snowav' + cfg.snowav_version,
              'data_type': '',
              'data_location': trdir,
              'file_type': '',
              'config_file': cfg.config_file,
              'proc_time': datetime.now()}

    insert(cfg.connector, 'RunMetadata', values)

    cfg.vid = {}
    for v in [*cfg.variables.variables.keys()]:
        u = cfg.variables.variables[v]['units']
        if (('vol' in v) or ('avail' in v)):
            u = cfg.units
        if (('z' in v) or (v == 'depth')):
            u = cfg.depthlbl
        if v == 'density':
            u = 'kg m^-3'
        if v == 'coldcont':
            u = 'MJ'

        variables = {'run_id': cfg.run_id,
                     'variable': v,
                     'unit': u,
                     'name': cfg.variables.variables[v]['description']}

        insert(cfg.connector, 'VariableUnits', variables)

        # After inserting into VariableUnits, get the existing id
        session = make_session(cfg.connector)
        qry = session.query(VariableUnits).filter(
            VariableUnits.run_id == cfg.run_id)
        session.close()
        df = pd.read_sql(qry.statement, qry.session.connection())
        cfg.vid[v] = df[df['variable'] == v]['id'].values[0]

    # snow_line
    variables = {'run_id': cfg.run_id,
                 'variable': 'snow_line',
                 'unit': cfg.elevlbl,
                 'name': 'snow_line'}

    insert(cfg.connector, 'VariableUnits', variables)
    session = make_session(cfg.connector)
    qry = session.query(VariableUnits).filter(
        VariableUnits.run_id == cfg.run_id)
    session.close()
    df = pd.read_sql(qry.statement, qry.session.connection())
    cfg.vid['snow_line'] = df['id'].values[0]


def connect(sqlite=None, sql=None, plotorder=None, user=None,
            password=None, host=None, port=None, convert=False,
            add=False):
    '''
    This establishes a connection with a database for results. If the specified
    sqlite database doesn't exist, it will be created.

    Note: consider moving creation of mysql database to Makefile

    Args
    --------
    sqlite : str
        path to sqlite database
    sql : str
        snowav mysql identifier
    plotorder : list
        list of basins
    user : str
        database user
    password : str
        database password
    host : int
        database host
    port : int
        database port

    Returns
    ---------
    basins : dict
        dictionary of snowav basins and ids for database connection
    logger : list

    '''

    logger = []

    if sqlite is not None:
        fp = urllib.parse.urlparse(sqlite)

        if (not os.path.isfile(fp.path)):
            database = sqlite
            logger.append(' Creating {} for results'.format(database))
            log = create_tables(database, plotorder)

            for out in log:
                logger.append(out)

            engine = create_engine(sqlite)

        else:
            engine = create_engine(sqlite)

        DBSession = sessionmaker(bind=engine)
        session = DBSession()
        logger.append(' Using {} for results'.format(sqlite))
        connector = sqlite

    if (sql is not None) and (sqlite is None):

        try:
            cnx = mysql.connector.connect(user=user,
                                          password=password,
                                          host=host,
                                          port=port)
        except:
            print('Failed attempting to make database connection with '
                  '{}:{}@{}/{}\nCheck config options in [database] '
                  'section'.format(user, password, host, port))
            exit()

        cursor = cnx.cursor()

        # Check if database exists, create if necessary
        query = ("SHOW DATABASES")
        cursor.execute(query)
        dbs = cursor.fetchall()

        try:
            dbs = [i[0].decode("utf-8") for i in dbs]
        except:
            dbs = [i[0] for i in dbs]

        db_engine = 'mysql+mysqlconnector://{}:{}@{}:{}/{}'.format(user,
                                                                   password,
                                                                   host,
                                                                   port,
                                                                   sql)

        # If the database doesn't exist, create it, otherwise connect
        if (sql not in dbs):
            logger.append(' Specified mysql database {} '.format(sql) +
                          'does not exist, it is being created...')
            query = ("CREATE DATABASE {};".format(sql))
            cursor.execute(query)
            log = create_tables(db_engine, plotorder)

            engine = create_engine(db_engine)
            Base.metadata.create_all(engine)
            DBSession = sessionmaker(bind=engine)
            session = DBSession()

            for out in log:
                logger.append(out)

        else:
            query = ("SHOW DATABASES")
            cursor.execute('USE {}'.format(sql))
            cursor.execute('SHOW TABLES')
            tbls = cursor.fetchall()

            # This hasn't been tested with new docker mysql instance, could be
            # a point of failure...
            if tbls is None:
                log = create_tables(db_engine, plotorder)

                for out in log:
                    logger.append(out)

                logger.append(log)

            try:
                logger.append(' Using database connection {}@{}:{} "{}" for '
                              'results'.format(user, host, port, sql))
                engine = create_engine(db_engine)
                Base.metadata.create_all(engine)
                DBSession = sessionmaker(bind=engine)
                session = DBSession()

            except:
                logger.append(' Failed trying to make database connection '
                              'to {}'.format(sql))

        cursor.close()
        cnx.close()

        connector = db_engine

    qry = session.query(Watershed)
    ws = pd.read_sql(qry.statement, qry.session.connection())

    qry = session.query(Basin)
    bs = pd.read_sql(qry.statement, qry.session.connection())

    if convert:
        plotorder[0] = convert_watershed_names(plotorder[0])

    if plotorder[0] in ws['watershed_name'].values:
        f = ws[ws['watershed_name'].values == plotorder[0]]
        wid = f['watershed_id'].values[0]

    else:
        logger.append(' {} not found as a watershed in '.format(plotorder[0]) +
                      'existing database, it is being added')

        if not ws.empty:
            wid = np.max(ws['watershed_id'].values) + 1
        else:
            wid = 1

        wval = Watershed(watershed_id=int(wid),
                         watershed_name=plotorder[0])
        if add:
            session.add(wval)

        else:
            logger.append(' Config option [database] add_basins: False, so '
                          '{} will not be added'.format(plotorder[0]))

            print('WARNING! Given current config options and topo.nc file, '
                  'trying to use {} as a watershed, but it does not currently '
                  'exist as a watershed in the database...\nCorrect [snowav] '
                  'masks or consider changing [database] add_basins: True '
                  'if you know what you are doing.'.format(plotorder[0]))

        session.commit()

    basins = {}
    watersheds = {plotorder[0]: {'watershed_id': wid, 'watershed_name': '',
                                 'basins': '', 'shapefile': ''}}

    # Initialize basins within the watershed
    for i, name in enumerate(plotorder):

        # This logic doesn't work if trying to 'move' a subbasin from one watershed to another,
        # i.e., 'Cherry Creek' going from part of the Extended Tuolumne to Tuolumne River
        # Basin in wy2020
        if name in bs['basin_name'].values:
            bid = int(bs[(bs['watershed_id'] == wid) & (bs['basin_name'] == name)]['basin_id'].values[0])

        else:
            logger.append(' {} not found as a basin in '.format(name) +
                          'existing database, it is being added')

            if not bs.empty:
                bid = np.max(bs['basin_id'].values) + 1
            else:
                bid = 1

            bval = Basin(watershed_id=int(wid),
                         basin_id=int(bid + i),
                         basin_name=name)

            if add:
                session.add(bval)

            else:
                logger.append(' Config option [database] add_basins: False, so '
                              '{} will not be added'.format(name))

                print('WARNING! Given current config options and topo.nc file, '
                      'trying to use {} as a basin, but it does not currently '
                      'exist as a basin in the database...\nConsider '
                      'changing [database] add_basins: True if you know what '
                      'you are doing.'.format(name))

        basins[name] = {'watershed_id': wid, 'basin_id': bid}

    session.commit()
    session.close()

    return basins, connector, logger


def convert_watershed_names(name):
    '''
    Convert watershed from topo.nc mask to existing snowav database version.

    '''
    convert_list = ['Boise River Basin', 'Lakes Basin', 'Merced River Basin',
                    'Kaweah River Basin', 'Kings River Basin']

    if name in convert_list:
        watershed = name.split(' ')[0]

    elif name == 'San Joaquin River Basin':
        watershed = 'San Joaquin'

    else:
        watershed = name

    return watershed


def package(connector, basins, df, run_id, vid, output, dtime):
    """ Put process() results on the database.

    Args
    ------
    connector {str}: database connector
    basins {dict}: snowav basins identifier
    run_id {int}: run_id for database
    vid {int}: variable id
    df {DataFrame}: results dataframe
    output {str}: output variable (i.e. 'swe_z')
    dtime {datetime}: datetime
    """

    for basin in df:

        for iters, val in enumerate(df[basin].values):
            if np.isnan(val):
                val = None
            else:
                val = float(val)

            values = {'basin_id': int(basins[basin]['basin_id']),
                      'run_id': int(run_id),
                      'date_time': dtime,
                      'variable': output,
                      'variable_id': int(vid[output]),
                      'value': val,
                      'elevation': str(df[basin].index[iters])}

            insert(connector, 'Results', values)
